#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level facilities for serializing and deserializing Numpy arrays to and
from comma-separated value (CSV) files.
'''

# ....................{ IMPORTS                           }....................
import numpy as np
from beartype import beartype
from beartype.typing import (
    Dict,
    Iterable,
    Optional,
)
from betse.exceptions import BetseSequenceException, BetseStrException
from betse.util.io.log import logs
from betse.util.path import dirs
from betse.util.type import types

# ....................{ WRITERS                           }....................
#FIXME: Donate this function back to Numpy as a new np.savecsv() function
#paralleling the existing np.savetxt() function.
@beartype
def write_csv(
    # Mandatory parameters.
    filename: str,
    column_name_to_values: Dict[str, Iterable],

    # Optional parameters.
    column_name_to_format: Optional[Dict[str, str]] = None,
) -> None:
    '''
    Serialize each key-value pair of the passed ordered dictionary into a new
    column in comma-separated value (CSV) format to the plaintext file with the
    passed filename.

    Caveats
    ----------
    To ensure that all columns of this file have the same number of rows, all
    values of this dictionary *must* be one-dimensional sequences of:

    * The same length. If this is not the case, an exception is raised.
    * Any type satisfying the :class:`SequenceTypes` API, including:

      * Numpy arrays.
      * Lists.
      * Tuples.

    Typically, each such value is a one-dimensional Numpy array of floats.

    Parameters
    ----------
    filename : str
        Absolute or relative path of the plaintext file to be written. If this
        file already exists, this file is silently overwritten.
    column_name_to_values: Dict[str, Iterable]
        Dictionary of all columns to be serialized such that:

        * Each key of this dictionary is a **column name** (i.e., terse string
          describing the type of data contained in this column).
        * Each value of this dictionary is **column data** (i.e.,
          one-dimensional sequence of all arbitrary data comprising this
          column).
    column_name_to_format : Optional[Dict[str, str]]
        Dictionary of all format strings formatting each column such that:

        * Each key of this dictionary is a **column name** (i.e., terse string
          describing the type of data contained in this column).
        * Each key of this dictionary is a **format string** (i.e.,
          ``%``-prefixed format string as fully documented by the "list of
          specifiers, one per column" subsection for the ``fmt`` parameter
          accepted by the low-level :func:`numpy.savetxt` function wrapped by
          this high-level writer function).

        This dictionary *must* have the exact same keys as the
        ``column_name_to_values`` dictionary.  Defaults to ``None``, in which
        case NumPy unconditionally defaults to reasonable floating-point format
        strings for *all* columns.

    Raises
    ----------
    BetseSequenceException
        If one or more values of this dictionary are either:

        * *Not* sequences.
        * Sequences whose length differs from that of any preceding value
          sequences of this dictionary.
    BetseStrException
        If this column name contains one or more characters reserved for use by
        the CSV non-standard, including:

        * Double quotes, reserved for use as the CSV quoting character.
        * Newlines, reserved for use as the CSV row delimiting character.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    # Log this serialization.
    logs.log_debug('Writing CSV file: %s', filename)

    # Tuple of all format strings to be passed to the numpy.savetxt() function
    # below (in column order), defaulting to the same format string as accepted
    # by that function for simplicity.
    columns_format = '%.18e'

    # If passed a dictionary of format strings...
    if column_name_to_format is not None:
        # If the column names listed by this dictionary from those listed by
        # the dictionary of values, raise an exception.
        if column_name_to_format.keys() != column_name_to_values.keys():
            raise BetseSequenceException(
                f'"column_name_to_format" keys '
                f'{repr(column_name_to_format.keys())} != '
                f'"column_name_to_values" keys '
                f'{repr(column_name_to_values.keys())} != '
            )
        # Else, these column names coincide.

        # Tuple of these format strings (in the same order).
        columns_format = tuple(
            column_format
            for column_format in column_name_to_format.values()
        )

    # Create the directory containing this file if needed and hence raise
    # filesystem-related exceptions *BEFORE* performing any subsequent logic.
    dirs.make_parent_unless_dir(filename)

    # Validate the contents of this dictionary. While the np.column_stack()
    # function called below also does, the exceptions raised by the latter are
    # both ambiguous and non-human-readable and hence effectively useless.
    #
    # Length of all prior columns *OR* "None" if no columns have been iterated.
    columns_prior_len = None

    # List of all column names sanitized such that each name containing one or
    # more comma characters is double-quoted.
    column_names = []

    # For each passed column...
    for column_name, column_values in column_name_to_values.items():
        # If this column is *NOT* a sequence, raise a human-readable exception.
        if not types.is_sequence_nonstr(column_values):
            raise BetseSequenceException(
                'Column "{}" type {!r} not a sequence.'.format(
                    column_name, type(column_values)))

        # Length of this column.
        column_len = len(column_values)

        # If this is the first column to be iterated, require all subsequent
        # columns be of the same length.
        if columns_prior_len is None:
            columns_prior_len = column_len
        # Else if this column's length differs from that of all prior columns,
        # raise a human-readable exception.
        elif column_len != columns_prior_len:
            raise BetseSequenceException(
                'Column "{}" length {} differs from '
                'length {} of prior columns.'.format(
                    column_name, column_len, columns_prior_len))

        # If this column name contains one or more reserved characters, raise
        # an exception. This includes:
        #
        # * Double quotes, reserved for use as the CSV quoting character.
        # * Newlines, reserved for use as the CSV row delimiting character.
        if '"' in column_name:
            raise BetseStrException(
                'Column name {} contains '
                "one or more reserved '\"' characters.".format(column_name))
        if '\n' in column_name:
            raise BetseStrException(
                'Column name {} contains '
                'one or more newline characters.'.format(column_name))

        # If this column name contains one or more commas (reserved for use as
        # the CSV delimiter), double-quote this name. Since the prior logic
        # guarantees this name to *NOT* contain double quotes, no further logic
        # is required
        if ',' in column_name:
            column_name = '"{}"'.format(column_name)

        # Append this sanitized column name to this list of such names.
        column_names.append(column_name)

    # Comma-separated string listing all column names.
    columns_name = strjoin.join_on(column_names, delimiter=',')

    # Two-dimensional sequence of all column data, whose:
    #
    # * First dimension indexes each column.
    # * Second dimension indexes each data point in that column.
    columns_values = tuple(column_name_to_values.values())

    # Two-dimensional Numpy array of all row data converted from this column
    # data, whose:
    #
    # * First dimension indexes each sampled time step such that each item is a
    #   one-dimensional Numpy array of length the number of columns (i.e., the
    #   number of key-value pairs in the passed dictionary).
    # * Second dimension indexes each column data point for that time step.
    #
    # Note that passing "column_name_to_values.values()" no longer suffices
    # under NumPy >= 1.16. Because breaking backward compatibility for utterly
    # no good reason is Now A Good Thing, the np.column_stack() function now
    # longer accepts arbitrary iterables but now requires strict sequences:
    #
    #     FutureWarning: arrays to stack must be passed as a "sequence" type
    #     such as list or tuple. Support for non-sequence iterables such as
    #     generators is deprecated as of NumPy 1.16 and will raise an error in
    #     the future.
    #
    # Converting non-sequence iterables to sequences is trivial, ignoring the
    # obvious edge case of infinite generators, which -- much like the
    # legendary Pokemon of yore -- appear not to exist in the wild. Ergo, the
    # NumPy developers yet again chose poorly. *STOP BREAKING EVERYTHING.*
    rows_values = np.column_stack(columns_values)

    # Serialize these sequences to this file in CSV format.
    np.savetxt(
        fname=filename,
        X=rows_values,
        fmt=columns_format,
        header=columns_name,
        delimiter=',',

        # Prevent Numpy from prefixing the above header by "# ". Most popular
        # software importing CSV files implicitly supports a comma-delimited
        # first line listing all column names.
        comments='',
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Linux-specific facilities.
'''

# ....................{ IMPORTS                           }....................
import platform
# from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import StrOrNoneTypes

# ....................{ TESTERS                           }....................
@func_cached
def is_linux() -> bool:
    '''
    ``True`` only if the current platform is either Linux or a platform
    masquerading to a reasonably accurate degree as Linux (e.g., the Windows
    Subsystem for Linux (WSL) but *not* Cygwin Windows, which is sufficiently
    different to warrant differentiation).
    '''

    return platform.system() == 'Linux'


@func_cached
def is_mir() -> bool:
    '''
    ``True`` only if the active Python interpreter is running under the
    Linux-specific Mir compositor, implying this process to be headfull and
    hence support both CLIs and GUIs.

    Caveats
    ----------
    For sanity, this function incorrectly assumes *all* Mir compositors to
    comply with the X Desktop Group (XDG) standard by exporting the
    ``${XDG_SESSION_TYPE}`` environment variable with a value of ``mir``. As
    this is *not* necessarily the case, this function may return false
    negatives for edge-case Mir compositors.

    See Also
    ----------
    :func:`_get_xdg_session_type_or_none`
        Further details.
    '''

    return _get_xdg_session_type_or_none() == 'mir'


@func_cached
def is_wayland() -> bool:
    '''
    ``True`` only if the active Python interpreter is running under a
    Linux-specific Wayland compositor, implying this process to be headfull and
    hence support both CLIs and GUIs.

    Caveats
    ----------
    For sanity, this function incorrectly assumes *all* Wayland compositors to
    comply with the X Desktop Group (XDG) standard by exporting the
    ``${XDG_SESSION_TYPE}`` environment variable with a value of ``wayland``.
    As this is *not* necessarily the case, this function may return false
    negatives for edge-case Wayland compositors.

    See Also
    ----------
    :func:`_get_xdg_session_type_or_none`
        Further details.
    '''

    return _get_xdg_session_type_or_none() == 'wayland'

# ....................{ GETTERS ~ distro                  }....................
@func_cached
def get_distro_name() -> str:
    '''
    Human-readable name of the current Linux distribution.

    Specifically, this function returns:

    * If the :func:`platform.linux_distribution` function is available, the
      first element of the 3-tuple returned by this function (e.g.,
      ``CentOS``).
    * Else, the string returned by the :func:`platform.system` function.
      Since this is probably the generic string ``Linux``, this is only a
      fallback.
    '''

    # Avoid circular import dependencies.
    from betse.lib import libs
    from betse.util.io.error.errwarning import ignoring_deprecations
    from betse.util.os.brand import windows

    # Linux distribution name to be returned.
    distro_name = None

    # If the deprecated Linux-specific platform.linux_distribution()
    # function is still defined...
    if hasattr(platform, 'linux_distribution'):
        # Ignore all deprecations emitted by Python 3.5 through 3.7 concerning
        # the pending removal of the platform.linux_distribution() function...
        with ignoring_deprecations():
            # Defer to the second item of the 3-tuple "(distname, version, id)"
            # Defer to the first item of the 3-tuple "(distname, version, id)"
            # returned by this function (e.g., "CentOS"). Since
            # platform.system() returns the low-level kernel name "Linux", the
            # latter name is ignored if feasible.
            distro_name = platform.linux_distribution()[0]
    # Else if the optional third-party Linux-specific "distro" package is
    # importable, defer to this package.
    elif libs.is_runtime_optional('distro'):
        # Import this package.
        distro = libs.import_runtime_optional('distro')

        # Defer to the name reported by this package.
        distro_name = distro.name()
    # Else if this is actually the Windows Subsystem for Linux (WSL)
    # masquerading as Linux, return "Windows (WSL)" rather than "Linux".
    elif windows.is_windows_wsl():
        distro_name = 'Windows (WSL)'
    # Else, this is Python >= 3.8 *AND* the "distro" package is unimportable.
    # In this case, default to the platform name returned by the
    # platform.system() function. Since this typically corresponds to the
    # low-level name of the current kernel (e.g., "Linux") rather than the
    # high-level name of the current Linux distribution (e.g., "Arch",
    # "Gentoo"), this is only a fallback of last resort.
    else:
        distro_name = platform.system()

    # Return this version specifier.
    return distro_name


@func_cached
def get_distro_version() -> str:
    '''
    Human-readable ``.``-delimited version specifier string of the current
    platform.

    Specifically, This function returns:

    * Under Linux, the version reported by this Linux distribution as follows:

      * If the deprecated :func:`platform.linux_distribution` function is still
        defined, the second element of the 3-tuple returned by this function
        (e.g., ``6.4``).
      * Else if the third-party Linux-specific "distro" package (referenced
    # by official Python documentation as a stand-in replacement for the
    # deprecated Linux-specific platform.linux_distribution() function) is
    # importable, defer to this package.
      * Else, the string returned by the :func:`platform.release` function.
        Since this is probably the non-human-readable ``.``- and
        ``-``-delimited version specifier string of the current Linux kernel
        (e.g., ``4.1.15-gentoo-r1``), this is only a fallback.
    '''

    # Avoid circular import dependencies.
    from betse.lib import libs
    from betse.util.io.error.errwarning import ignoring_deprecations

    # Version specifier to be returned.
    distro_version = None

    # If the deprecated Linux-specific platform.linux_distribution()
    # function is still defined...
    if hasattr(platform, 'linux_distribution'):
        # Ignore all deprecations emitted by Python 3.5 through 3.7 concerning
        # the pending removal of the platform.linux_distribution() function...
        with ignoring_deprecations():
            # Defer to the second item of the 3-tuple "(distname, version, id)"
            # returned by this function (e.g., "6.4"). Since platform.release()
            # returns the low-level kernel version (e.g., "4.1.15"), the latter
            # version is ignored if feasible.
            distro_version = platform.linux_distribution()[1]
    # Else if the optional third-party Linux-specific "distro" package is
    # importable, defer to this package.
    elif libs.is_runtime_optional('distro'):
        # Import this package.
        distro = libs.import_runtime_optional('distro')

        # For disambiguity, prefer the "best" (i.e., most specific and
        # accurate) version specifier published by this Linux distribution.
        distro_version = distro.version(best=True)
    # Else, this is Python >= 3.8 *AND* the "distro" package is unimportable.
    # In this case, default to the version specifier returned by the
    # platform.release() function. Since this typically corresponds to the
    # low-level version of the current kernel (e.g., "4.1.15") rather than the
    # high-level version of the current Linux distribution (e.g., "6.4"), this
    # is a fallback of last resort.
    else:
        distro_version = platform.release()

    # Return this version specifier.
    return distro_version

# ....................{ PRIVATE ~ getters                 }....................
@func_cached
def _get_xdg_session_type_or_none() -> StrOrNoneTypes:
    '''
    Lowercase machine-readable string standardized by the X Desktop Group (XDG)
    identifying the general category of Linux-specific display manager under
    which the active Python interpreter is running if any *or* ``None``
    otherwise (e.g., if this interpreter is running headless).

    Returns
    ----------
    StrOrNoneTypes
        Either:

        * If either the current platform is not Linux *or* the current shell
          environment does not define the ``${XDG_SESSION_TYPE}`` variable
          queried by this function, ``None``. In the latter case, this commonly
          but *not* necessarily implies this interpreter to be running
          headless.
        * Else, the value of the ``${XDG_SESSION_TYPE}`` variable, guaranteed
          to be one of the following strings:

            * If this interpreter is running headless under an interactive
              pseudoterminal, ``tty``.
            * If this interpreter is running headfull under:

              * The Mir compositor, ``mir``.
              * A Wayland compositor, ``wayland``.
              * The X Window System, ``x11``.

            * If something has gone non-deterministic, ``unspecified``.

    See Also
    ----------
    https://www.freedesktop.org/software/systemd/man/sd_session_get_type.html
        Low-level PAM-specific ``sd_session_get_type(3)`` function retrieving
        the same string.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.shell import shellenv

    # If the current platform is *NOT* Linux, return "None".
    if not is_linux():
        return None
    # Else, the current platform is Linux.

    # Return the string value of the ${XDG_SESSION_TYPE} environment variable
    # if defined *OR* "None" otherwise.
    return shellenv.get_var_or_none('XDG_SESSION_TYPE')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for :mod:`pkg_resources`, a mandatory runtime
dependency simplifying inspection of application dependencies.
'''

#FIXME: *EXCISE THIS ENTIRE SUBMODULE, PLEASE.* This should no longer be
#required, due to importing deprecated functionality.

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseLibException
from betse.util.io.log import logs
from betse.util.type.types import (
    type_check,
    # DistributionOrNoneTypes,
    GeneratorType,
    MappingType,
    ModuleType,
    ModuleOrSequenceTypes,
    SequenceTypes,
)
from collections import OrderedDict
from setuptools import pkg_resources
from setuptools.pkg_resources import (
    DistributionNotFound,
    Requirement,
    UnknownExtra,
    VersionConflict,
)

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_requirements_dict(requirements_dict: MappingType) -> None:
    '''
    Raise an exception unless each :mod:`setuptools`-formatted requirements
    string produced by concatenating each key and corresponding value of the
    passed dictionary (e.g., converting key ``Numpy`` and value ``>= 1.8.0``
    into requirements string ``Numpy >= 1.8.0``) are satisfiable, implying
    these third-party packages to be both importable and of satisfactory
    version.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings to validate.

    Raises
    ----------
    BetseLibException
        If at least one such requirement is unsatisfiable.
    '''

    # Tuple of requirements strings converted by concatenating each key and
    # value of this dictionary.
    requirements_str = get_requirements_str_from_dict(requirements_dict)

    # Validate these requirements strings.
    die_unless_requirements_str(*requirements_str)


@type_check
def die_unless_requirements_dict_keys(
    requirements_dict: MappingType, *requirement_names: str) -> None:
    '''
    Raise an exception unless each :mod:`setuptools`-formatted requirements
    string produced by concatenating each passed key and corresponding value of
    the passed dictionary (e.g., converting key ``Numpy`` and value ``>=
    1.8.0`` into requirements string ``Numpy >= 1.8.0``) are satisfiable,
    implying these third-party packages to be both importable and of
    satisfactory version.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings to validate a subset of.
    requirement_names : Tuple[str]
        Tuple of keys identifying the key-value pairs of this dictionary to
        validate.

    Raises
    ----------
    BetseLibException
        If at least one such requirement is unsatisfiable.
    '''

    # Tuple of requirements strings converted by concatenating each such key
    # and value of this dictionary.
    requirements_str = get_requirements_str_from_dict_keys(
        requirements_dict, *requirement_names)

    # Validate these requirements strings.
    die_unless_requirements_str(*requirements_str)


@type_check
def die_unless_requirements_str(*requirements_str: str) -> None:
    '''
    Raise an exception unless *all* passed :mod:`setuptools`-formatted
    requirements strings (e.g., ``Numpy >= 1.8.0``) are satisfiable, implying
    the corresponding third-party packages to be both importable and of
    satisfactory version.

    See Also
    ----------
    :func:`is_requirement_str`
        Further details.

    Raises
    ----------
    BetseLibException
        If at least one such requirement is unsatisfiable.
    '''

    # List of all requirement objects parsed from these requirement strings.
    requirements = iter_requirements_str(*requirements_str)

    # Validate each such requirement.
    for requirement in requirements:
        die_unless_requirement(requirement)


@type_check
def die_unless_requirement(requirement: Requirement) -> None:
    '''
    Raise an exception unless the passed :mod:`setuptools`-specific requirement
    is satisfiable, implying the corresponding third-party module or package to
    be both importable and of satisfactory version.

    Parameters
    ----------
    requirement : Requirement
        :mod:`setuptools`-specific object describing this module or package's
        requisite name and version.

    Raises
    ----------
    BetseLibException
        If this requirement is unsatisfiable.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodule
    from betse.util.py.module.pymodname import (
        DEPENDENCY_TO_MODULE_NAME)

    # Human-readable exception to be raised below if any.
    betse_exception = None

    # Human-readable name of this module or package.
    requirement_name = requirement.project_name

    # Fully-qualified name of this requirement's module or package.
    package_name = DEPENDENCY_TO_MODULE_NAME[requirement_name]

    # Attempt to manually import this requirement's module or package.
    try:
        package = import_requirement(requirement)
    # If a standard import exception is raised...
    except ImportError as root_exception:
        # Low-level Python-specific exception message.
        root_exception_message = str(root_exception)

        # If this exception signifies the common case of a missing dependency,
        # avoid exposing this exception to end users. Doing so would convey no
        # meaningful metadata.
        if root_exception_message == (
            "No module named '{}'".format(package_name)):
            betse_exception = BetseLibException(
                'Dependency "{}" not found.'.format(requirement_name))
        # Else, this exception signifies an unexpected edge-case. For
        # debuggability, expose this exception to end users.
        else:
            raise BetseLibException(
                'Dependency "{}" unimportable.'.format(requirement_name))
    # Else if any other exception is raised, expose this exception as is.
    except Exception:
        raise BetseLibException(
            'Dependency "{}" unimportable.'.format(requirement_name))

    # If raising a human-readable exception, do so. Raising this exception in
    # the handler above would be preferable but also incite Python 3 to
    # implicitly prepend this exception with the non-human-readable exception
    # raised above. This convoluted logic circumvents that. (...Ugh!)
    if betse_exception:
        raise betse_exception

    # If this requirement is unversioned, all possible versions of this package
    # satisfy this requirement, in which case this requirement is satisfied.
    if not _is_requirement_versioned(requirement):
        return
    # Else, this requirement is versioned.

    # Package version if any *OR* "None" otherwise.
    package_version = pymodule.get_version_or_none(package)

    # If this package declares a version...
    if package_version is not None:
        # If this version satisfies this requirement,  we're done here.
        if package_version in requirement:
            return
        # Else, this version fails to satisfy this requirement. In this case,
        # raise an exception.
        else:
            raise BetseLibException(
                'Dependency "{}" unsatisfied by installed version {}.'.format(
                    requirement, package_version))
    # Else, this package declares *NO* version. In this case, fallback to
    # unreliable setuptools-specific logic.

    # Attempt to...
    try:
        # Setuptools-specific object describing the current version of the
        # package satisfying this requirement if any *OR* "None" if this
        # requirement cannot be guaranteed to be unsatisfied.
        distribution = get_requirement_distribution_or_none(requirement)

        # If this requirement is satisfied, we're done here.
        if distribution is not None:
            return
        # Else, this requirement is unsatisfied. In this case, raise a generic
        # non-descript exception.
        else:
            raise BetseLibException(
                'Dependency "{}" unsatisfied.'.format(requirement))
    # If setuptools only found requirements of insufficient version, a
    # non-human-readable exception resembling the following is raised:
    #
    #    pkg_resources.VersionConflict: (PyYAML 3.09 (/usr/lib64/python3.3/site-packages), Requirement.parse('PyYAML>=3.10'))
    #
    # Detect this and raise a human-readable exception instead.
    except VersionConflict as version_conflict:
        betse_exception = BetseLibException(
            'Dependency "{}" unsatisfied by '
            'installed dependency "{}".'.format(
                version_conflict.req, version_conflict.dist))
    #FIXME: Handle the "UnknownExtra" exception as well.

    # If raising a human-readable exception, do so.
    if betse_exception:
        raise betse_exception
    # Else, this requirement is satisfied.

# ....................{ TESTERS                           }....................
@type_check
def is_requirement_str(*requirements_str: str) -> bool:
    '''
    ``True`` only if all passed :mod:`setuptools`-formatted requirement strings
    (e.g., `Numpy >= 1.8.0`) are satisfiable, implying the corresponding
    third-party packages to be importable *and* of satisfactory versions.

    Parameters
    ----------
    requirements_str : tuple[str]
        Tuple of all requirement strings to test.

    Returns
    ----------
    bool
        ``True`` only if these requirements are all satisfiable.
    '''

    # List of all requirement objects parsed from these requirement strings.
    requirements = iter_requirements_str(*requirements_str)

    # Return "True" only if these requirements are all satisfiable.
    return all(
        is_requirement(requirement)
        for requirement in requirements
    )


@type_check
def is_requirement(requirement: Requirement) -> bool:
    '''
    ``True`` only if the passed :mod:`setuptools`-specific requirement is
    satisfiable, implying the corresponding third-party module or package to be
    both importable and of satisfactory version.

    Parameters
    ----------
    requirement : Requirement
        :mod:`setuptools`-specific object describing this module or package's
        requisite name and version.

    Returns
    ----------
    bool
        ``True`` only if this requirement is satisfiable.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodule

    # Attempt to manually import this requirement's package.
    try:
        package = import_requirement(requirement)
    # If this package is unimportable, reduce this exception to a boolean.
    except ImportError:
        return False
    # If any other exception is raised, expose this exception as is.

    # If this requirement is unversioned, all possible versions of this package
    # satisfy this requirement, in which case this requirement is satisfied.
    if not _is_requirement_versioned(requirement):
        return True
    # Else, this requirement is versioned.

    # Package version if any *OR* "None" otherwise.
    package_version = pymodule.get_version_or_none(package)

    # If this package declares a version...
    if package_version is not None:
        # Return true only if this version satisfies this requirement.
        return package_version in requirement
    # Else, this package declares *NO* version. In this case, fallback to
    # unreliable setuptools-specific logic.

    # Attempt to...
    try:
        # Setuptools-specific object describing the current version of the
        # package satisfying this requirement if any *OR* "None" if this
        # requirement cannot be guaranteed to be unsatisfied.
        distribution = get_requirement_distribution_or_none(requirement)

        # Return true only if this requirement is satisfied.
        return distribution is not None
    # If setuptools found only requirements of insufficient version, reduce
    # this exception to a boolean.
    except (UnknownExtra, VersionConflict):
        return False
    # If any other exception is raised, expose this exception as is.

# ....................{ TESTERS ~ private                 }....................
@type_check
def _is_requirement_versioned(requirement: Requirement) -> bool:
    '''
    ``True`` only if the passed :mod:`setuptools`-specific requirement is
    **versioned** (i.e., constrained to require only a subset of all available
    versions of this requirement's distribution).

    Parameters
    ----------
    requirement : Requirement
        Object describing this module or package's required name and version.

    Returns
    ----------
    bool
        ``True`` only if this requirement is versioned.
    '''

    # The "Requirement.specifier" instance variable of type "SpecifierSet"
    # encapsulates the set of all versions required by this requirement. Note:
    #
    # * The SpecifierSet.__init__() method accepts a string defining this set
    #   (e.g., ">= 3.1415"), which is the empty string when no specific
    #   versions are required, in which case this set is empty.
    # * The SpecifierSet.__eq__() method permits such sets to compared against
    #   such strings.
    #
    # Together, these two methods imply that a requirement is versioned if and
    # only if this requirement's specifier is *NOT* the empty string.
    return requirement.specifier != ''

# ....................{ GETTERS ~ requirement              }....................
@type_check
def get_requirement_distribution_or_none(requirement: Requirement) -> object:  # DistributionOrNoneTypes:
    '''
    :class:`Distribution` instance describing the currently installed version
    of the top-level third-party module or package satisfying the passed
    :mod:`setuptools`-specific requirement if any *or* raise an exception if
    this requirement is guaranteed to be unsatisfied (e.g., due to a version
    mismatch) *or* ``None`` if this requirement cannot be guaranteed to be
    unsatisfied (e.g., due to this requirement being installed either without
    :mod:`setuptools` or with the :mod:`setuptools` subcommand ``develop``).

    Caveats
    ----------
    **Callers are advised to call this getter only as a last-ditch fallback.**
    The :mod:`setuptools`-agnostic ``pip`` and ``pip3`` commands rather than
    the :mod:`setuptools`-specific ``easy_install`` command are now commonly
    used to install Python 3 applications. For unknown reasons, the former
    occasionally fail to update :mod:`setuptools`-specific distribution
    metadata, which then become desynchronized from the underlying packages
    described by those distributions. Instead, consider dynamically importing
    and directly accessing attributes exposed by those packages (e.g., the
    standard ``__init__.__version__`` package attribute).

    That said, this high-level getter should *always* be called in lieu of the
    low-level :func:`pkg_resources.get_distribution` function, which raises
    spurious exceptions in common non-erroneous edge cases (e.g., packages
    installed via the :mod:`setuptools` subcommand ``develop``) and is thus
    unsafe for general-purpose use.

    Parameters
    ----------
    requirement : Requirement
        Object describing this module or package's required name and version.

    Returns
    -------
    DistributionOrNoneTypes
        Object describing the currently installed version of the package or
        module satisfying this requirement if any *or* ``None`` otherwise.
        Specifically, ``None`` is returned in all of the following conditions
        -- only one of which genuinely corresponds to an error:

        * This requirement is *not* installed at all. (**Error.**)
        * This requirement was installed manually rather than with
          :mod:`setuptools`, in which case no such :class:`Distribution`
          exists. (**Non-error.**)
        * This requirement was installed with the :mod:`setuptools` subcommand
          ``develop``, in which case a :class:`Distribution` technically exists
          but in a sufficiently inconsistent state that the low-level
          :func:`pkg_resources.get_distribution` function raises an exception
          on attempting to retrieve that object. (**Non-error.**)

        Since distinguishing the erroneous from non-erroneous cases exceeds the
        mandate of this getter, the caller is expected to do so.

    Raises
    ----------
    VersionConflict
        If the currently installed version of this module or package fails to
        satisfy this requirement's version constraints.
    '''

    # If setuptools finds this requirement, return its distribution as is.
    try:
        # logs.log_debug(
        #     'Retrieving requirement "%r" distribution...', requirement)
        return pkg_resources.get_distribution(requirement)
    # If setuptools fails to find this requirement, this does *NOT* necessarily
    # imply this requirement to be unimportable as a package. Rather, this only
    # implies this requirement was *NOT* installed as a setuptools-managed egg.
    # This requirement is still installable and hence importable (e.g., by
    # manually copying this requirement's package into the "site-packages"
    # subdirectory of the top-level directory for this Python interpreter).
    # However, does this edge-case actually occur in reality? *YES.*
    # PyInstaller-frozen applications embed requirements without corresponding
    # setuptools-managed eggs. Hence, this edge-case *MUST* be handled.
    except DistributionNotFound:
        # logs.log_debug(
        #     'Requirement "%r" distribution not found.', requirement)
        return None
    # If setuptools fails to find the distribution-packaged version of this
    # requirement (e.g., due to having been editably installed with "sudo
    # python3 setup.py develop"), this version may still be manually parseable
    # from this requirement's package. Since setuptools fails to raise an
    # exception whose type is unique to this error condition, the contents of
    # this exception are parsed to distinguish this expected error condition
    # from other unexpected error conditions. In the former case, a
    # non-human-readable exception resembling the following is raised:
    #
    #     ValueError: ("Missing 'Version:' header and/or PKG-INFO file",
    #     networkx [unknown version] (/home/leycec/py/networkx))
    except ValueError as version_missing:
        # If this exception was...
        if (
            # ...instantiated with two arguments...
            len(version_missing.args) == 2 and
            # ...whose second argument is suffixed by a string indicating the
            # version of this distribution to have been ignored rather than
            # recorded during installation...
            str(version_missing.args[1]).endswith(' [unknown version]')
        # ...this exception indicates an expected ignorable error condition.
        # Silently ignore this edge case.
        ):
            # logs.log_debug(
            #     'Requirement "%r" distribution version not found.', requirement)
            return None

        # Else, this exception indicates an unexpected and hence
        # non-ignorable error condition. Reraise this exception!
        raise


#FIXME: Rename to convert_requirements_dict_key_to_str() for parity with the
#BETSEE codebase.
@type_check
def get_requirement_str_from_dict_key(
    requirements_dict: MappingType, requirement_name: str) -> str:
    '''
    Convert the key-value pair of the passed dictionary of
    :mod:`setuptools`-specific requirements strings whose key is the passed
    string into a :mod:`setuptools`-specific requirements string.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings.
    requirement_names : str
        Key identifying the key-value pairs of this dictionary to convert.

    Returns
    ----------
    str
        Requirements string converted from this key-value pair.

    Raises
    ----------
    BetseLibException
        If the passed key is *not* a key of this dictionary.

    See Also
    ----------
    :func:`get_requirements_str_from_dict`
        Further details on the format of this dictionary and resulting string.
    '''

    # If this name is unrecognized, raise an exception.
    if requirement_name not in requirements_dict:
        raise BetseLibException(
            'Dependency "{}" unrecognized.'.format(requirement_name))

    # String constraining this requirement (e.g., ">= 3.14159265359", "[svg]")
    # if any or "None" otherwise.
    requirement_constraints = requirements_dict[requirement_name]

    # If this requirement is constrained, return the concatenation of this
    # requirement's name and constraints.
    if requirement_constraints:
        return '{} {}'.format(requirement_name, requirement_constraints)
    # Else, return only this requirement's name.
    else:
        return requirement_name


@type_check
def get_requirement_synopsis(requirement: Requirement) -> str:
    '''
    Human-readable string describing the currently installed third-party
    module or package corresponding to (but *not* necessarily satisfying) the
    passed :mod:`setuptools`-specific requirement.

    This function is principally intended for use in printing package metadata
    in a non-critical manner and hence is guaranteed to *never* raise fatal
    exceptions. If this module or package:

    * Is unimportable, the string ``not installed`` is returned.
    * Is importable and...

      * Fails to satisfy this requirement, a string describing this conflict is
        returned.
      * Satisfies this requirement, the string ``{version} <{pathname}>`` is
        returned, where:

        * ``{pathname}`` is the absolute path of this module or package.
        * ``{version}`` is either:

          * If this module or package exports a version (e.g., via the
            ``__version__`` attribute), this version as a `.`-delimited string.
          * Else, the string ``unknown version``.

    Parameters
    ----------
    requirement : Requirement
        Object describing this module or package's required name and version.

    Returns
    ----------
    str
        Human-readable string describing this requirement.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodule

    # Distribution satisfying this requirement if any or "None" otherwise.
    distribution = None

    try:
        # Object describing the currently installed version of the package or
        # module satisfying this requirement if any or "None" if this
        # requirement cannot be guaranteed to be unsatisfied.
        distribution = get_requirement_distribution_or_none(requirement)
    # If setuptools found only requirements of insufficient version, return
    # this version regardless (with a suffix noting this to be the case).
    except VersionConflict as version_conflict:
        return '{} [fails to satisfy {}]'.format(
            version_conflict.dist.version, version_conflict.req)
    #FIXME: Handle the "UnknownExtra" exception as well.

    # Attempt to manually import this module or package.
    try:
        package = import_requirement(requirement)
    # If this package is unimportable, return a human-readable string.
    except ImportError:
        return 'not installed'

    # Pathname and version of this module or package.
    package_pathname = pymodule.get_filename(package)
    package_version = None

    # If this requirement is satisfied, reuse the version provided by this
    # requirement's high-level distribution -- which is guaranteed to be at
    # least as precise as version metadata provided by this requirement's
    # low-level module or package (if any).
    #
    # Unfortunately, although the "Distribution" class does provide a public
    # "location" instance variable, this variable's value is typically the
    # absolute path of the parent directory containing this module or package
    # rather than the absolute path of the latter. The former is overly
    # ambiguous and hence useless for our purposes. We have no recourse but to
    # manually import this module or package and inspect its attributes.
    if distribution is not None:
        package_version = distribution.version
    # Else, this requirement is unsatisfied. Fallback to the version metadata
    # provided by this requirement's low-level module or package.
    else:
        package_version = pymodule.get_version_or_none(package)

        # If no such version is provided, default to a human-readable string.
        if package_version is None:
            package_version = 'unknown version'

    # Return the expected string in the event of success.
    return '{} <{}>'.format(package_version, package_pathname)

# ....................{ GETTERS ~ requirements : dict     }....................
@type_check
def get_requirements_dict_from_str(
    requirements_tuple: SequenceTypes) -> dict:
    '''
    Convert the passed tuple of :mod:`setuptools`-specific requirements strings
    into a dictionary of such strings.

    This tuple is assumed to contain strings of the form
    ``{project_name} {project_specs}``, where:

    * ``{project_name}`` is the :mod:`setuptools`-specific project name of a
      third-party dependency (e.g., ``NetworkX``).
    * ``{project_specs}`` is a :mod:`setuptools`-specific requirements string
      constraining this dependency (e.g., ``>= 1.11``).

    Each key-value pair of the resulting dictionary maps from each such
    ``{project_name}`` to the corresponding ``{project_specs}``.

    Parameters
    ----------
    requirements_tuple : SequenceTypes
        Tuple of :mod:`setuptools`-specific requirements strings in the format
        described above.

    Returns
    ----------
    dict
        Dictionary of :mod:`setuptools`-specific requirements strings in the
        format described above.
    '''

    # List of all requirement objects parsed from these requirement strings.
    requirements = iter_requirements_str(*requirements_tuple)

    # Dictionary containing these requirements.
    requirements_dict = {}

    # For each such requirement...
    for requirement in requirements:
        # Comma-delimited string of all constraints of this requirement
        # iteratively reduced from the high-level "specs" sequence of 2-tuples
        # "(op, version)" of this requirement object into low-level strings.
        #
        # Technically, manually parsing each requirement string of the passed
        # tuple into the project name and specifications required below would
        # also be feasible. Since manual parsing is significantly more fragile
        # than deferring to the authoritative parsing already implemented by
        # the canonical "pkg_resources" module, the latter is preferred.
        requirement_specs_str = ','.join(
            '{} {}'.format(requirement_spec[0], requirement_spec[1])
            for requirement_spec in requirement.specs
        )

        # Add a new key-value pair to this dictionary mapping from the name of
        # this requirement to the comma-delimited string of all constraints of
        # this requirement.
        requirements_dict[requirement.project_name] = requirement_specs_str

    # Return this dictionary.
    return requirements_dict

# ....................{ GETTERS ~ requirements : str      }....................
#FIXME: Rename to convert_requirements_dict_to_tuple() for parity with the
#BETSEE codebase.
@type_check
def get_requirements_str_from_dict(requirements_dict: MappingType) -> tuple:
    '''
    Convert the passed dictionary of :mod:`setuptools`-specific requirements
    strings into a tuple of such strings.

    This dictionary is assumed to map from the :mod:`setuptools`-specific
    project name of a third-party dependency (e.g., ``NetworkX``) to the suffix
    of a :mod:`setuptools`-specific requirements string constraining this
    dependency (e.g., ``>= 1.11``). Each element of the resulting tuple is a
    string of the form `{key} {value}`, converted from a key-value pair of this
    dictionary in arbitrary order.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of :mod:`setuptools`-specific requirements strings in the
        format described above.

    Returns
    ----------
    tuple
        Tuple of :mod:`setuptools`-specific requirements strings in the format
        described above.
    '''

    return get_requirements_str_from_dict_keys(
        requirements_dict, *requirements_dict.keys())


#FIXME: Rename to convert_requirements_dict_keys_to_tuple() for parity with the
#BETSEE codebase. This function is painfully misnamed as is.
@type_check
def get_requirements_str_from_dict_keys(
    requirements_dict: MappingType, *requirement_names: str) -> tuple:
    '''
    Convert all key-value pairs of the passed dictionary of
    :mod:`setuptools`-specific requirements strings whose keys are the passed
    strings into a tuple of :mod:`setuptools`-specific requirements strings.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings.
    requirement_names : Tuple[str]
        Tuple of keys identifying the key-value pairs of this dictionary to
        convert.

    Returns
    ----------
    tuple
        Tuple of :mod:`setuptools`-specific requirements strings in the above
        format.

    Raises
    ----------
    BetseLibException
        If the passed key is *not* a key of this dictionary.

    See Also
    ----------
    :func:`get_requirements_str_from_dict`
        Further details on the format of this dictionary and resulting strings.
    '''

    return tuple(
        get_requirement_str_from_dict_key(
            requirements_dict, requirement_name)
        for requirement_name in requirement_names
    )

# ....................{ GETTERS ~ requirements : synopsis }....................
@type_check
def get_requirements_dict_synopsis(
    requirements_dict: MappingType) -> OrderedDict:
    '''
    Ordered dictionary synopsizing the currently installed third-party packages
    corresponding to (but *not* necessarily satisfying) the
    :mod:`setuptools`-formatted requirements strings produced by concatenating
    each key and value of the passed dictionary (e.g., converting key ``Numpy``
    and value ``>= 1.8.0`` into requirements string ``Numpy >= 1.8.0``).

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings to retrieve metadata for.

    Returns
    ----------
    OrderedDict
        Ordered dictionary lexicographically presorted on keys in ascending
        order such that each:

        * Key is a string of the form ``{{requirement_name}} version``.
        * Value is either:

          * The currently installed version specifier of this requirement
            if this requirement is both installed and exports an
            inspectable version at runtime (e.g., an ``__version__``
            package attribute).
          * The string ``not installed`` if this requirement is not found.
          * The string ``unknown version`` if this requirement exports no
            inspectable version at runtime.
    '''

    # Tuple of requirements strings converted by concatenating each key and
    # value of this dictionary.
    requirements_str = get_requirements_str_from_dict(requirements_dict)

    # Return this metadata.
    return get_requirements_str_synopsis(*requirements_str)


@type_check
def get_requirements_str_synopsis(*requirements_str: str) -> OrderedDict:
    '''
    Ordered dictionary synopsizing the currently installed third-party packages
    corresponding to (but *not* necessarily satisfying) the passed
    :mod:`setuptools`-formatted requirements strings (e.g., ``Numpy >= 1.8``).

    Parameters
    ----------
    requirements_str : Tuple[str]
        Tuple of requirements strings to retrieve metadata for.

    Returns
    ----------
    OrderedDict
        Ordered dictionary lexicographically presorted on keys in ascending
        order (as detailed by the :func:`get_requirements_dict_synopsis`
        function).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import itersort

    # Lexicographically sorted tuple of these strings.
    requirement_strs_sorted = itersort.sort_ascending(requirements_str)

    # List of all requirement objects parsed from these requirement strings.
    requirements = iter_requirements_str(*requirement_strs_sorted)

    # Ordered dictionary synopsizing these requirements.
    return OrderedDict(
        (requirement.project_name, get_requirement_synopsis(requirement))
        for requirement in requirements)

# ....................{ IMPORTERS                         }....................
@type_check
def import_requirements_dict_keys(
    requirements_dict: MappingType, *requirements_name: str) -> (
    ModuleOrSequenceTypes):
    '''
    Import and return either a single top-level module object if passed one
    :mod:`setuptools`-specific requirement name *or* sequence of top-level
    module objects otherwise (i.e., if passed multiple such names).

    Each returned module object is guaranteed to satisfy the
    :mod:`setuptools`-specific requirement string produced by concatenating
    the passed requirement name and value of the passed dictionary whose key
    is that name (e.g., converting key ``six`` and value ``>= 7.9`` into the
    requirements string ``six >= 7.9``).

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary of requirements strings to import and yield a subset of.
    requirements_name : tuple[str]
        Tuple of keys identifying the key-value pairs of this dictionary to be
        imported and yielded.

    Returns
    ----------
    ModuleOrSequenceTypes
        Either:

        * If only one requirement name was passed, the top-level module object
          satisfying the corresponding requirement of this dictionary.
        * Else, a tuple of top-level module objects satisfying the
          corresponding requirements of this dictionary.

    See Also
    ----------
    :func:`import_requirements_str`
        Further details.
    '''

    # Tuple of requirements strings converted by concatenating each such key
    # and value of this dictionary.
    requirements_str = get_requirements_str_from_dict_keys(
        requirements_dict, *requirements_name)

    # Import and return every top-level module for these requirements strings.
    return import_requirements_str(*requirements_str)


@type_check
def import_requirements_str(*requirements_str: str) -> ModuleOrSequenceTypes:
    '''
    Import and return either a single top-level module object if passed one
    :mod:`setuptools`-specific requirement strings (e.g., ``six >= 7.9``) *or*
    a sequence of top-level module objects otherwise (i.e., if passed multiple
    such strings).

    Each returned module object is guaranteed to satisfy the corresponding
    :mod:`setuptools`-specific requirement string.

    Parameters
    ----------
    requirements_str : tuple[str]
        Tuple of all such requirements strings to import and return modules of.

    Returns
    ----------
    ModuleOrSequenceTypes
        Either:

        * If only one requirement name was passed, the top-level module object
          satisfying the corresponding requirement of this dictionary.
        * Else, a tuple of top-level module objects satisfying the
          corresponding requirements of this dictionary.

    See Also
    ----------
    :func:`import_requirement`
        Further details.
    '''

    # List of all requirement objects parsed from these requirement strings.
    requirements = iter_requirements_str(*requirements_str)

    # Tuple of all modules imported from these requirement objects.
    requirements_module = tuple(
        import_requirement(requirement) for requirement in requirements)

    # Return:
    #
    # * If more than one such module was imported, this tuple of these modules.
    # * Else, the single module imported above.
    return (
        requirements_module if len(requirements_module) > 1 else
        requirements_module[0])


@type_check
def import_requirement(requirement: Requirement) -> ModuleType:
    '''
    Import and return the top-level module object satisfying the passed
    :mod:`setuptools`-specific requirement.

    Parameters
    ----------
    requirement : Requirement
        Object describing this module or package's required name and version.

    Returns
    ----------
    ModuleType
        Top-level package object implementing this requirement.

    Raises
    ----------
    ImportError
        If this package is unimportable.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodname
    from betse.util.py.module.pymodname import (
        DEPENDENCY_TO_MODULE_NAME)

    # Fully-qualified name of this requirement's package.
    package_name = DEPENDENCY_TO_MODULE_NAME[
        requirement.project_name]

    # Log this importation, which can often have unexpected side effects.
    logs.log_debug('Importing third-party package "%s"...', package_name)

    # Import and return this package.
    return pymodname.import_module(package_name)

# ....................{ ITERATORS                         }....................
@type_check
def iter_requirements_str(*requirements_str: str) -> GeneratorType:
    '''
    Generator iteratively yielding the high-level :mod:`setuptools`-specific
    :class:`pkg_resources.Requirement` objects parsed from each passed
    low-level requirement string.

    Parameters
    ----------
    requirements_str: Tuple[str]
        Tuple of all requirement strings to yield requirement objects for.

    Yields
    ----------
    Requirement
        Requirement parsed from each passed requirement string (in the passed
        order).
    '''

    yield from pkg_resources.parse_requirements(requirements_str)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **request** (i.e., the standard ``request`` fixture exposed by the
:mod:`pytest` API) functionality.

This submodule principally defines utility functions simplifying access to this
fixture. The official documentation for this fixture is blatantly inadequate.
Even were that *not* the case, however, this fixture's API is overly obtuse to
an absurd (and almost obfuscatory) degree. This submodule defines an
intelligible alternative to those horrors.

Caveats
-------
All functions defined below accepting a ``request`` fixture object assume the
passed object to be of the desired type rather than validating this to be the
case. Why? Because that type resides in a private module that does *not*
appear to be externally importable, thanks to devilish import trickery: e.g.,

    >>> import _pytest.python.FixtureRequest as req
    ImportError: cannot import name 'transfer_markers'

See Also
--------
https://pytest.org/latest/builtin.html#_pytest.python.FixtureRequest
    Official documentation for this fixture – such as it is.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseTestFixtureException
from betse.util.test.pytest import pytests
from betse.util.type.iterable import sequences
from betse.util.type.types import type_check, TestableTypes
from pytest import Function

# ....................{ CONSTANTS                          }....................
_DEFAULT_VALUE_NONE = object()
'''
Default value for the optional ``default_value`` argument accepted by the
:func:`get_fixture_param` function.

This value permits that function to distinguish between default values of value
``None`` and the lack of any default value altogether.
'''

# ....................{ EXCEPTIONS                         }....................
def die_unless_tested(request) -> None:
    '''
    Raise an exception unless the passed ``request`` fixture object was
    transitively requested by a test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Raises
    ------
    BetseTestFixtureException
        If no test transitively requested this ``request`` fixture.

    See Also
    --------
    :func:`.is_tested`
        Further details.
    '''

    if not is_tested(request):
        raise BetseTestFixtureException(
            '"request" fixture transitively requested by no test '
            '(e.g., due to being requested by a file- or '
            'session-scoped fixture): {}'.format(request))

# ....................{ EXCEPTIONS ~ fixture               }....................
def die_unless_fixture(request) -> None:
    '''
    Raise an exception unless the passed ``request`` fixture object was
    directly requested by a fixture rather than a test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Raises
    ------
    BetseTestFixtureException
        If a test rather than fixture requested this ``request`` fixture.
    '''

    if not is_fixture(request):
        raise BetseTestFixtureException(
            '"request" fixture requested by '
            'test rather than fixture: {}'.format(request))


def die_unless_fixture_parametrized(request) -> bool:
    '''
    Raise an exception unless the fixture requesting the passed ``request``
    fixture object was parametrized (either directly or indirectly).

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Raises
    ------
    BetseTestFixtureException
        If either:
        * A test rather than fixture requested this ``request`` fixture.
        * A fixture that was _not_ parametrized requested this ``request``
          fixture.
    '''

    if not is_fixture_parametrized(request):
        raise BetseTestFixtureException(
            '"request" fixture requested by '
            'unparametrized fixture "{}".'.format(get_fixture_name(request)))

# ....................{ TESTERS                            }....................
def is_tested(request) -> bool:
    '''
    ``True`` only if the passed ``request`` fixture object was transitively
    requested by a test (e.g., directly by a test or indirectly by a test-scope
    fixture).

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    -------
    bool
        ``True`` only if a test transitively requested this ``request``
        fixture.
    '''

    return isinstance(request.node, Function)

# ....................{ TESTERS ~ fixture                  }....................
def is_fixture(request) -> bool:
    '''
    ``True`` only if the passed ``request`` fixture object was directly
    requested by a fixture rather than a test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    bool
        ``True`` only if a fixture requested this ``request`` fixture.
    '''

    return request.fixturename is not None


def is_fixture_parametrized(request) -> bool:
    '''
    ``True`` only if the fixture requesting the passed ``request`` fixture
    object was parametrized (either directly or indirectly).

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    -------
    bool
        ``True`` only if this fixture was parametrized.

    Raises
    ------
    BetseTestFixtureException
        If a test rather than fixture requested this ``request`` fixture.
    '''

    # If this "request" fixture was requested by a test, raise an exception.
    die_unless_fixture(request)

    # Else, that fixture was itself requested by a fixture. Return whether on
    # not the latter was parametrized. py.test API: you fail again.
    return hasattr(request, 'param')

# ....................{ GETTERS                            }....................
def get_tested_name(request) -> str:
    '''
    Unqualified name of the current test (e.g., ``test_cli_info``) if the
    passed ``request`` fixture object was either directly requested by that
    test *or* indirectly requested by a test-scope fixture.

    This function returns either:

    * If the current test is a test function, this function's unqualified name.
    * If the current test is a test method, this method's unqualified name.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    -------
    str
        Unqualified name of this test.

    Raises
    ------
    BetseTestFixtureException
        If the passed ``request`` fixture object was *not* transitively
        requested by a test (e.g., was requested by a file- or session-scope
        fixture), in which case no test exists to obtain the name of.
    '''

    # Raise an exception unless a test transitively requested this fixture.
    die_unless_tested(request)

    # No one would ever think of trying this. Except one man did.
    return request.node.name

# ....................{ GETTERS ~ fixture                  }....................
def get_fixture_name(request) -> str:
    '''
    Name of the current fixture that explicitly requested the passed
    ``request`` fixture object.

    If this ``request`` fixture was requested by a test rather than a fixture,
    an exception is raised.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    -------
    str
        Name of such fixture.
    '''

    # If this "request" fixture was requested by a test, raise an exception.
    die_unless_fixture(request)

    # Else, that fixture was itself requested by a fixture. Return the name of
    # the latter.
    return request.fixturename


@type_check
def get_fixture_param(
    request,
    default_value: object = _DEFAULT_VALUE_NONE,
    type_expected: TestableTypes = None,
) -> object:
    '''
    Parameter parametrizing the fixture requesting the passed ``request``
    fixture, optionally defaulted to the passed default value if the former
    fixture is unparametrized *and* optinionally validated to be an instance of
    the passed type or tuple of types.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    default_value: optional[object]
        Default value to return if this fixture is unparametrized. Defaults
        to :data:`_DEFAULT_VALUE_NONE`, in which case an exception is raised
        instead if this fixture is unparametrized.
    type_expected : optional[TestableTypes]
        Class or tuple of classes to validate this object to be an instance of.
        If this object is *not* an instance of this class or classes, an
        exception is raised. Defaults to ``None``, in which case this object is
        returned unvalidated.

    Returns
    -------
    object
        Parameter parametrizing this fixture.

    Raises
    ------
    BetseTestFixtureException
        If either:

        * A test rather than fixture requested this ``request`` fixture.
        * An unparametrized fixture requested this ``request`` fixture.
        * A fixture parametrized by an instance of a class *not* the passed
          ``check_type`` requested this ``request`` fixture.
    '''

    # Parameter parametrizing the fixture requesting this "request" fixture
    fixture_param = None

    # If the caller passed a default value *AND* this fixture is
    # unparametrized, default this parameter to this value.
    if (default_value is not _DEFAULT_VALUE_NONE and
        not is_fixture_parametrized(request)):
        fixture_param = default_value
    # Else...
    else:
        # Raise an exception unless this fixture is parametrized.
        die_unless_fixture_parametrized(request)

        # Parameter parametrizing this fixture.
        fixture_param = request.param

    # Else, that fixture was parametrized. If the caller requested that
    # parametrization be validated, do so.
    if type_expected is not None and not isinstance(
        fixture_param, type_expected):
        raise BetseTestFixtureException(
            'Fixture "{}" parametrization not of type "{}": {!r}'.format(
                get_fixture_name(request), type_expected, fixture_param))

    # Return this parameter.
    return fixture_param

# ....................{ GETTERS ~ request                  }....................
#FIXME: This function internally calls the now-deprecated
#request.getfuncargvalue() function, which has since been replaced by the
#request.getfixturevalue() function. Sadly, the latter is unavailable in the
#most recent stable version of "pytest" available under my Linux distribution
#(Gentoo for the win) and hence ignored for the moment.
#
#Revisit this in early 2017, please.

@type_check
def get_requested_fixture(
    request,
    fixture_name: str,
    type_expected: TestableTypes = None,
) -> object:
    '''
    Object returned by the fixture with the passed name transitively requested
    by the current test requesting the passed ``request`` fixture object,
    optinionally validated to be an instance of the passed type or tuple of
    types.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name: str
        Name of the fixture to return the object returned by that fixture.
    type_expected : optional[TestableTypes]
        Class or tuple of classes to validate this object to be an instance of.
        If this object is *not* an instance of this class or classes, an
        exception is raised. Defaults to ``None``, in which case this object is
        returned unvalidated.

    Returns
    -------
    _pytest.python.FixtureDef
        Object returned by this fixture.

    Raises
    ------
    :func:`pytests.get_pytest_fixture_lookup_error_type`
        If this fixture is either unavailable *or* is available but
        unretrievable (e.g., this fixture is the child of the current fixture).
    '''

    # Object returned by this fixture. Terrible Function Names Part XIXI: the
    # endless journey endures.
    fixture_object = request.getfuncargvalue(fixture_name)

    # If the caller requested this object be validated, do so.
    if type_expected is not None and not isinstance(
        fixture_object, type_expected):
        raise BetseTestFixtureException(
            'Fixture "{}" object not of type "{}": {!r}'.format(
                fixture_name, type_expected, fixture_object))

    # Return this object.
    return fixture_object


def get_requested_fixture_or_none(request, fixture_name: str) -> object:
    '''
    Fixture with the passed name transitively requested by the current test if
    this fixture is both available and retrievable *or* ``None`` otherwise,
    inspected from the passed ``request`` fixture object.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name: str
        Name of the fixture to return.

    Returns
    -------
    {_pytest.python.FixtureDef, None}
        This fixture if this fixture is both available and retrievable *or*
        ``None`` otherwise.
    '''

    # Class of all "FixtureLookupError" exceptions raised by "pytest".
    FixtureLookupError = pytests.get_pytest_fixture_lookup_error_type()

    # Attempt to retrieve and return this fixture for this request.
    try:
        return get_requested_fixture(request, fixture_name)
    # If no such fixture exists, return None.
    except FixtureLookupError:
        return None

# ....................{ GETTERS ~ request : name           }....................
def get_requested_fixture_names(request) -> list:
    '''
    List of the names of all fixtures transitively requested by the current
    test excluding the fixture that explicitly requested the passed ``request``
    fixture object if any, inspected from that ``request`` fixture object.

    This list includes the names of all fixtures that are:

    * Directly requested by this test.
    * Directly or indirectly requested by fixtures requested by this test.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.

    Returns
    ----------
    list
        List of all such names.
    '''

    # List of the names of all fixtures transitively required by this test,
    # including the fixture that explicitly requested this "request" fixture if
    # any. Since the original list is guaranteed *NOT* to be modified below,
    # this list need not be copied from the original list to preserve the
    # latter.
    #
    # This attribute is undocumented. Why is this attribute undocumented? Work
    # with me here, people.
    fixture_names = request.fixturenames

    # If a fixture explicitly requested this "request" fixture...
    if is_fixture(request):
        # Name of this fixture.
        omit_fixture_name = get_fixture_name(request)

        # Exclude this name from this list *WITHOUT* modifying the original
        # list, as doing so would subsequently raise "KeyError" exceptions.
        fixture_names = sequences.omit_item(fixture_names, omit_fixture_name)
        assert omit_fixture_name not in fixture_names

    # Return these fixture names.
    return fixture_names

# ....................{ GETTERS ~ request : name : prefix  }....................
def get_requested_fixture_name_prefixed_by(
    request, fixture_name_prefix: str) -> str:
    '''
    Name prefixed by the passed prefix of the single fixture transitively
    requested by the current test, inspected from the passed ``request``
    fixture object.

    If either no such fixture *or* more than one such fixture exist, an
    exception is raised.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name_prefix: str
        String prefixing the fixture name to be returned.

    Returns
    -------
    str
        Such name.

    See Also
    --------
    :func:`.get_requested_fixture_names`
        Further details.
    '''

    # List of the names of all fixtures prefixed by this prefix.
    prefixed_fixture_names = get_requested_fixture_names_prefixed_by(
        request, fixture_name_prefix)

    # Number of such fixtures.
    prefixed_fixture_count = len(prefixed_fixture_names)

    # If either no or more than one such fixtures exist, raise an exception.
    if prefixed_fixture_count != 1:
        # Exception message to be raised.
        exception_message = None

        if prefixed_fixture_count == 0:
            exception_message = (
                'No fixture prefixed by "{}" requested by this test.'.format(
                    fixture_name_prefix))
        else:
            exception_message = (
                'Multiple fixtures prefixed by "{}" '
                'requested by this test: {}'.format(
                    fixture_name_prefix, prefixed_fixture_names))

        # Raise this exception with this message.
        raise BetseTestFixtureException(exception_message)

    # Else, return the single such fixture name.
    return prefixed_fixture_names[0]


def get_requested_fixture_names_prefixed_by(
    request, fixture_name_prefix: str) -> list:
    '''
    List of all names prefixed by the passed prefix of all fixtures
    transitively requested by the current test, inspected from the passed
    ``request`` fixture object.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Object passed to fixtures and tests requesting the ``request`` fixture.
    fixture_name_prefix: str
        String prefixing all fixture names in the returned list.

    Returns
    -------
    list
        List of all such names.

    See Also
    --------
    :func:`.get_requested_fixture_names`
        Further details.
    '''

    return sequences.get_items_prefixed_by(
        sequence=get_requested_fixture_names(request),
        item_prefix=fixture_name_prefix,
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Implementation of the ``info`` subcommand for this command line interface
(CLI).
'''

#FIXME; For aesthetics, convert to yppy-style "cli.memory_table" output.

# ..................{ IMPORTS                               }..................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid non-trivial delays on importing this module, this module
# should import *ONLY* from modules and packages whose importation is unlikely
# to incur such delays. This includes all standard Python packages and all
# application packages required by the log_info_header() function, which is
# called sufficiently early in application startup as to render these imports
# effectively mandatory.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse import metadata
from betse.util.io.log import logs
from betse.util.io.log.conf import logconf
from betse.util.os import oses
from betse.util.py import pyimpl, pys
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict
from collections import OrderedDict

# ..................{ GETTERS                               }..................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current installation of this
    application.
    '''

    # Defer heavyweight imports.
    from betse.util.app.meta import appmetaone
    from betse.util.path import pathnames
    from betse.util.os.command import cmds

    # Return this dictionary.
    return OrderedArgsDict(
        'basename', cmds.get_current_basename(),
        'version',  metadata.VERSION,
        'codename', metadata.CODENAME,
        'authors',  metadata.AUTHORS,
        'license',  metadata.LICENSE,
        'home directory', pathnames.get_home_dirname(),
        'dot directory',  appmetaone.get_app_meta().dot_dirname,
        'data directory', appmetaone.get_app_meta().data_dirname,
    )

# ..................{ LOGGERS                               }..................
def log_header() -> None:
    '''
    Log a single-line human-readable sentence synopsizing the state of the
    current application (e.g., name, codename, version).
    '''

    logs.log_info(
        'Welcome to <<'
        '{program_name} {program_version} | '
        '{py_name} {py_version} | '
        '{os_name} {os_version}'
        '>>.'.format(
            program_name=metadata.NAME,
            program_version=metadata.VERSION,
            py_name=pyimpl.get_name(),
            py_version=pys.get_version(),
            os_name=oses.get_name(),
            os_version=oses.get_version(),
        ))


def log_info() -> None:
    '''
    Log all metadata required by the ``info`` subcommand.
    '''

    # Notify the current user of a possible wait *BEFORE* importing modules
    # whose importation contributes to this wait.
    logs.log_info('Harvesting system metadata... (This may take a moment.)')

    # Defer heavyweight imports.
    from betse.lib import libs
    from betse.util.os import displays, kernels

    #FIXME: Shift into a more appropriate general-purpose submodule.
    # Tuple of BETSE-specific metadata.
    BETSE_METADATAS = (
        # Application metadata.
        (metadata.NAME.lower(), get_metadata()),

        # Logging metadata.
        ('logging', logconf.get_metadata()),
    )

    #FIXME: Shift into a more appropriate general-purpose submodule.
    # Tuple of system-specific metadata.
    SYSTEM_METADATAS = (
        # Python metadata.
        ('python', pys.get_metadata()),
        ('python interpreter', pyimpl.get_metadata()),

        # Operating system (OS) metadata.
        ('os', oses.get_metadata()),
        ('os kernel', kernels.get_metadata()),
        ('os display', displays.get_metadata()),
    )

    # Dictionary of human-readable labels to dictionaries of all
    # human-readable keys and values categorized by such labels. All such
    # dictionaries are ordered so as to preserve order in output.
    info_type_to_dict = OrderedDict(
        BETSE_METADATAS +
        libs.get_metadatas() +
        SYSTEM_METADATAS
    )

    # String formatting this information.
    info_buffer = 'Harvested system information:\n'

    # Format each such dictionary under its categorizing label.
    for info_type, info_dict in info_type_to_dict.items():
        # Format this label.
        info_buffer += '\n{}:'.format(info_type)

        # Format this label's dictionary.
        for info_key, info_value in info_dict.items():
            info_buffer += '\n  {}: {}'.format(info_key, info_value)

    # Log rather than merely output this string, as logging simplifies
    # cliest-side bug reporting.
    logs.log_info(info_buffer)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-string iterable iterators** (i.e., utility functions converting
non-string objects implementing the abstract :class:`collections.abc.Iterable`
base class into similar objects for iteration purposes) facilities.
'''

# ....................{ IMPORTS                           }....................
import itertools
# from betse.util.io.log import logs
from betse.util.type.types import type_check, IterableTypes

# ....................{ ITERATORS ~ combinations          }....................
@type_check
def iter_pairs(iterable: IterableTypes) -> IterableTypes:
    '''
    Iterable of all possible pairs of items in the passed iterable.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.

    Returns
    ----------
    IterableTypes
        Iterable of all possible pairs of items in the passed iterable.

    See Also
    ----------
    :func:`iter_combinations`
        Further details.
    '''

    # Before the one-liner, there was only shared misery.
    return iter_combinations(iterable=iterable, k=2)


@type_check
def iter_combinations(iterable: IterableTypes, k: int) -> IterableTypes:
    '''
    Iterable of all possible **k-combinations** (i.e., subiterables of ``k``
    distinct items) of the passed iterable.

    The iterable created and returned by this function is guaranteed to contain
    exactly ``(n k)`` tuples of ``k`` items each, where:

    * ``n`` is the length of the passed iterable.
    * ``k`` is the passed parameter ``k``.
    * ``(n k)`` is the binomial coefficient -- a non-negative integer either:

      * If ``k > n``, 0.
      * If ``k <= n``, ``n!/[k!(n-k)!]``.

    Moreover, the order of both the k-combinations of this iterable *and* the
    items of each such k-combination is guaranteed to be the exact same as that
    of the passed iterable. Ergo, passing a sorted iterable returns a similarly
    sorted iterable.

    Motivation
    ----------
    This function is a thin wrapper about the standard C-based
    :func:`itertools.combinations` function, whose official documentation and
    examples decidedly leave something to be desired.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    k : int
        Length of each k-combination in the iterable to be returned. If this
        length is negative, an exception is raised.

    Returns
    ----------
    IterableTypes
        Iterable of all possible k-combinations of the passed iterable.

    Raises
    ----------
    ValueError
        If ``k < 0``.

    Examples
    ----------
        >>> from betse.util.type.iterable import iteriter
        >>> letter_to_woodburn_harris = (
        ...     'I love to dream, ',
        ...     'but I never try to dream ',
        ...     'and think at the same time.')
        >>> a_lover_of_dreams = iteriter.iter_combinations(
        ...     iterable=letter_to_woodburn_harris, k=2)
        >>> tuple(a_lover_of_dreams)
        (('I love to dream, ', 'but I never try to dream '),
         ('I love to dream, ', 'and think at the same time.'),
         ('but I never try to dream ', 'and think at the same time.'))
    '''

    # And then there was one... liners.
    return itertools.combinations(iterable, k)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.util.math.mathoper` submodule.
'''

# ....................{ TESTS                              }....................
def test_det1d() -> None:
    '''
    Unit test the :func:`betse.util.math.mathoper.det1d` function.
    '''

    # Defer heavyweight imports.
    from betse.util.math.mathoper import det1d
    from numpy import asarray
    from numpy.testing import assert_equal

    # Two-dimensional input arrays.
    arr1 = asarray([1, 2])
    arr2 = asarray([4, 5])

    # One-dimensional cross product of these arrays.
    determinant = det1d(arr1, arr2)

    # Assert this cross product is as expected.
    assert_equal(determinant, -3, strict=True)


def test_cross2d() -> None:
    '''
    Unit test the :func:`betse.util.math.mathoper.cross2d` function.
    '''

    # Defer heavyweight imports.
    from betse.util.math.mathoper import cross2d
    from numpy import asarray
    from numpy.testing import assert_equal

    # Two-dimensional input arrays.
    arr1 = asarray([[1, 2], [6, 5]])
    arr2 = asarray([[7, 8], [3, 4]])

    # One-dimensional cross product of these arrays.
    cross_product = cross2d(arr1, arr2)

    # Assert this cross product is as expected.
    assert_equal(cross_product, asarray([-6, 9]), strict=True)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Metadata constants synopsizing high-level application behaviour.

Design
----------
Metadata constants defined by this submodule are intentionally *not* defined as
metadata properties of the :class:`betse.util.app.meta.appmetaabc` abstract base
class. Why? Because doing so would prevent their use from the top-level
``setup.py`` scripts defined by downstream consumers (e.g., BETSEE GUI), which
would render these constants effectively useless for their principal use case.
'''

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: This submodule has largely (but *NOT* entirely) been obsoleted by the
# top-level "pyproject.toml" file, which should be strongly preferred and
# towards which we should refactor this project away from this submodule.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys

# ....................{ METADATA                           }....................
NAME = 'BETSE'
'''
Human-readable application name.
'''


LICENSE = '2-clause BSD'
'''
Human-readable name of the license this application is licensed under.
'''

# ....................{ PYTHON ~ version                   }....................
PYTHON_VERSION_MIN = '3.11.0'
'''
Human-readable minimum version of Python required by this application as a
``.``-delimited string.
'''


PYTHON_VERSION_MINOR_MAX = 14
'''
Maximum minor stable version of this major version of Python currently released
(e.g., ``5`` if Python 3.5 is the most recent stable version of Python 3.x).
'''


def _convert_version_str_to_tuple(version_str: str) -> tuple:
    '''
    Convert the passed human-readable ``.``-delimited version string into a
    machine-readable version tuple of corresponding integers.
    '''
    assert isinstance(version_str, str), (
        '"{}" not a version string.'.format(version_str))

    return tuple(int(version_part) for version_part in version_str.split('.'))


PYTHON_VERSION_MIN_PARTS = _convert_version_str_to_tuple(PYTHON_VERSION_MIN)
'''
Machine-readable minimum version of Python required by this application as a
tuple of integers.
'''


# Validate the version of the active Python interpreter *BEFORE* subsequent
# code possibly depending on this version. Since this version should be
# validated both at setuptools-based install time and post-install runtime
# *AND* since this module is imported sufficiently early by both, stash this
# validation here to avoid duplication of this logic and hence the hardcoded
# Python version.
#
# The "sys" module exposes three version-related constants for this purpose:
#
# * "hexversion", an integer intended to be specified in an obscure (albeit
#   both efficient and dependable) hexadecimal format: e.g.,
#    >>> sys.hexversion
#    33883376
#    >>> '%x' % sys.hexversion
#    '20504f0'
# * "version", a human-readable string: e.g.,
#    >>> sys.version
#    2.5.2 (r252:60911, Jul 31 2008, 17:28:52)
#    [GCC 4.2.3 (Ubuntu 4.2.3-2ubuntu7)]
# * "version_info", a tuple of three or more integers *OR* strings: e.g.,
#    >>> sys.version_info
#    (2, 5, 2, 'final', 0)
#
# For sanity, this application will *NEVER* conditionally depend upon the
# string-formatted release type of the current Python version exposed via the
# fourth element of the "version_info" tuple. Since the first three elements of
# that tuple are guaranteed to be integers *AND* since a comparable 3-tuple of
# integers is declared above, comparing the former and latter yield the
# simplest and most reliable Python version test.
#
# Note that the nearly decade-old and officially accepted PEP 345 proposed a
# new field "requires_python" configured via a key-value pair passed to the
# call to setup() in "setup.py" (e.g., "requires_python = ['>=2.2.1'],"), that
# field has yet to be integrated into either disutils or setuputils. Hence,
# that field is validated manually in the typical way.
if sys.version_info[:3] < PYTHON_VERSION_MIN_PARTS:
    # Human-readable current version of Python. Ideally, "sys.version" would be
    # leveraged here instead; sadly, that string embeds significantly more than
    # merely a version and hence is inapplicable for real-world usage: e.g.,
    #
    #     >>> import sys
    #     >>> sys.version
    #     '3.6.5 (default, Oct 28 2018, 19:51:39) \n[GCC 7.3.0]'
    PYTHON_VERSION = '.'.join(
        str(version_part) for version_part in sys.version_info[:3])

    # Die ignominiously.
    raise RuntimeError(
        '{} requires at least Python {}, but the active interpreter '
        'is only Python {}. We feel deep sadness for you.'.format(
            NAME, PYTHON_VERSION_MIN, PYTHON_VERSION))

# ....................{ METADATA ~ version                 }....................
VERSION = '1.4.1'
'''
Human-readable application version as a ``.``-delimited string.
'''


CODENAME = 'Nicest Nestor'
'''
Human-readable code name associated with the current version of this
application.

This code name consists of an arbitrary adjective followed by the last name of
an arbitrary academic associated with field of bioelectricity whose first
letter is the same as the first letter of that adjective.

See Also
----------
:download:`/doc/rst/RELEASE.md`
    Further details on this incredible code name schema.
'''


VERSION_PARTS = _convert_version_str_to_tuple(VERSION)
'''
Machine-readable application version as a tuple of integers.
'''

# ....................{ METADATA ~ tag                     }....................
GIT_TAG_COMPAT_OLDEST = 'v0.5.0'
'''
Git-specific tag of the oldest version of this application for which the
current version of this application guarantees backward compatibility.

In this context, the term "backward compatibility" means the capacity for all
versions of this application newer than this tag to successfully load all:

* Simulation configuration files loadable by this older version.
* Pickled seeds, initializations, and simulations saved by this older version.

This specific version was selected for the simple reason that this was the
first version of this application to guarantee backward compatibility, as
implemented by the :mod:`betse.science.compat.compatconf` submodule.

See Also
----------
:func:`betse_test.func.sim.test_sim.test_cli_sim_compat`
    Functional test in our test suite programmatically guaranteeing backward
    compatibility with this older version.
'''

# ....................{ METADATA ~ synopsis                }....................
# Note that a human-readable multiline description is exposed via the top-level
# "setup.py" script. This description is inefficiently culled from the contents
# of the top-level "README.rst" file and hence omitted here. (Doing so here
# would significantly increase program startup costs with little to no benefit.)
SYNOPSIS = 'BETSE, the BioElectric Tissue Simulation Engine.'
'''
Human-readable single-line synopsis of this application.

By PyPI design, this string must *not* span multiple lines or paragraphs.
'''


DESCRIPTION = (
    'The BioElectric Tissue Simulation Engine (BETSE) is a '
    'discrete exterior calculus simulator for '
    '2D computational multiphysics problems in '
    'the life sciences -- including '
    '(electro)diffusion, '
    '(electro)osmosis, '
    'galvanotaxis, '
    'voltage-gated ion channels, '
    'gene regulatory networks, '
    'and biochemical reaction networks.'
)
'''
Human-readable multiline description of this application.

By :mod:`argparse` design, this string may (and typically should) span both
multiple lines and paragraphs. Note that this string is *not* published to
PyPI, which accepts reStructuredText (rst) and is thus passed the contents of
the top-level :doc:`/README` file instead.
'''

# ....................{ METADATA ~ authors                 }....................
AUTHORS = 'Alexis Pietak, Cecil Curry, et al.'
'''
Human-readable list of all principal authors of this application as a
comma-delimited string.

For brevity, this string *only* lists authors explicitly assigned copyrights.
For the list of all contributors regardless of copyright assignment or
attribution, see the top-level `AUTHORS.md` file.
'''


AUTHOR_EMAIL = 'alexis.pietak@gmail.com'
'''
Email address of the principal corresponding author (i.e., the principal author
responding to public correspondence).
'''

# ....................{ METADATA ~ urls                    }....................
URL_HOMEPAGE = 'https://github.com/betsee/betse'
'''
URL of this application's homepage.
'''


URL_DOWNLOAD = f'{URL_HOMEPAGE}/archive/{VERSION}.tar.gz'
'''
URL of the source tarball for the current version of this application.

This URL assumes a tag whose name is ``v{VERSION}`` where ``{VERSION}`` is the
human-readable current version of this application (e.g., ``v0.4.0``) to exist.
Typically, no such tag exists for live versions of this application -- which
have yet to be stabilized and hence tagged. Hence, this URL is typically valid
*only* for previously released (rather than live) versions of this application.
'''


URL_ISSUES = f'{URL_HOMEPAGE}/issues'
'''
URL of this package's issue tracker.
'''


URL_RELEASES = f'{URL_HOMEPAGE}/releases'
'''
URL of this package's release list.
'''

# ....................{ METADATA ~ python                  }....................
PACKAGE_NAME = NAME.lower()
'''
Fully-qualified name of the top-level Python package implementing this
application.

Caveats
----------
**Prefer the application-agnostic
:meth:`betse.appmeta.app_meta.package_name` property instead.** This
application-specific global fails to generalize to downstream consumers (e.g.,
BETSEE) and hence is usable *ONLY* for low-level installation-time use cases.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **decorator-based memoization** (i.e., efficient caching and reuse of
the values returned by decorated callables) facilities.
'''

#FIXME: Generalize the @func_cached decorator into a new method_cached()
#decorator whose implementation will resemble (and possibly even be
#effectively identical to) the existing property_cached() decorator. After
#doing so, revise the docstring for the func_cached() function to remove all
#reference to bound methods.
#FIXME: After defining method_cached(), refactor this submodule as follows:
#
#* Create a new callable_cached() decorator, internally calling either
#  func_cached() or method_cached() conditionally depending on whether the
#  passed callable is a bound method or not.
#* Rename all existing "@func_cached" decorations to "@callable_cached".
#* Renome func_cached() to _func_cached.
#* Renome method_cached() to _method_cached.

# ....................{ IMPORTS                            }....................
from betse.util.type.types import type_check, CallableTypes, PropertyType
from functools import wraps

if False: wraps  # silence contemptible IDE warnings

# ....................{ CONSTANTS                          }....................
CALLABLE_CACHED_VAR_NAME_PREFIX = '_betse_cached__'
'''
Substring prefixing the names of all private instance variables to which all
caching decorators (e.g., :func:`property_cached`) cache values returned by
decorated callables.

This prefix guarantees uniqueness across *all* instances -- including those
instantiated from official Python and unofficial third-party classes and those
internally defined by this application. Doing so permits logic elsewhere (e.g.,
pickling filtering) to uniquely match and act upon these variables.
'''


FUNCTION_CACHED_VAR_NAME = f'{CALLABLE_CACHED_VAR_NAME_PREFIX}function_value'
'''
Name of the private instance variable to which the :func:`func_cached`
decorator statically caches the value returned by the decorated function.
'''


PROPERTY_CACHED_VAR_NAME_PREFIX = f'{CALLABLE_CACHED_VAR_NAME_PREFIX}property_'
'''
Substring prefixing the names of all private instance variables to which the
:func:`property_cached` decorator dynamically caches the value returned by the
decorated property method.
'''

# ....................{ DECORATORS                         }....................
#FIXME: Raise an exception if the passed callable is a bound method. Note that
#we have *NO* means of reliably testing for unbound methods, which are
#indistinguishable from non-class functions. *ALL* attempts to differentiate
#the two fail under one or more edge cases. See also these failed attempts to
#do just that at this StackOverflow question:
#    https://stackoverflow.com/questions/8793233/python-can-a-decorator-determine-if-a-function-is-being-defined-inside-a-class
@type_check
def func_cached(func: CallableTypes) -> CallableTypes:
    '''
    Decorate the passed **non-property callable** (e.g., function, lambda,
    method) to cache the value returned by the first call of this callable.

    On the first call of a callable decorated with this decorator, the passed
    callable is called with all passed parameters, the value returned by this
    callable is cached into a private attribute of this callable, and this
    value is returned. On each subsequent call of this callable, the cached
    value is returned as is *without* calling this callable. Hence, this
    callable is called at most once for each instance of the class containing
    this property.

    Caveats (Memoization)
    ----------
    **This decorator does not memoize callables.** Memoization would map each
    permutation of parameters passed to the decorated callable to a unique
    return value conditionally cached (and hence returned) for that
    permutation.

    This decorator instead unconditionally caches (and hence returns) a single
    return value for *all* permutations of passed parameters. Hence, this
    decorator is principally intended to decorate callables accepting *no*
    parameters (e.g., simple testers and getters).

    Caveats (Methods)
    ----------
    **This decorator is only intended to cache return values of functions.**
    This decorator is *not* intended to cache return values of non-function
    callables (e.g., bound or unbound methods). Ergo, this decorator is named
    ``func_cached`` rather than the more general-purpose name
    ``callable_cached``. While this decorator *could* technically be used to
    decorate non-function callables, it never should be. Why? Because:

    * Decorating lambdas is highly non-trivial and hence impractical.
    * Attempting to cache return values of unbound methods with this decorator
      would do so globally rather than on an instance-specific basis and hence
      be functionally useless. To cache return values of bound methods on an
      instance-specific basis, an entirely new decorator would need to be
      created -- presumably returning a descriptor whose ``__get__()`` special
      method creates and returns bound methods whose return values are cached
      onto instance variables of those bound methods themselves. For similar
      functionality, see the
      ``betse.util.type.descriptor.expralias.expr_alias`` data descriptor.

    See Also
    ----------
    :func:`property_cached`
        Decorator caching object properties.
    '''

    # Raw string of Python statements comprising the body of this wrapper.
    # See @property_cached for further details, including time profilings.
    #
    # Technically, note that this decorator *SHOULD* raise a
    # "BetseCallableException" if any parameter accepted by this callable has
    # the reserved name `__callable_cached`. However, validating this
    # constraint would require inspecting this callable's signature and
    # iteratively searching the parameter list of that signature for parameter
    # names violating this constraint. Since the likelihood of a callable
    # accepting such parameters is "None", doing so would increase time
    # complexity *WITHOUT* yielding tangible benefits. (Also, we're lazy.)
    func_body = '''
@wraps(__callable_cached)
def callable_cached_arged(
    *args, __callable_cached=__callable_cached, **kwargs):
    try:
        return __callable_cached.{callable_var_name}
    except AttributeError:
        __callable_cached.{callable_var_name} = __callable_cached(
            *args, **kwargs)
        return __callable_cached.{callable_var_name}
'''.format(callable_var_name=FUNCTION_CACHED_VAR_NAME)

    # Dictionary mapping from local attribute names to values. For efficiency,
    # only attributes required by the body of this wrapper are copied from the
    # current namespace. (See below.)
    local_attrs = {'__callable_cached': func}

    # Dynamically define this wrapper as a closure of this decorator. For
    # obscure and presumably uninteresting reasons, Python fails to locally
    # declare this closure when the locals() dictionary is passed; to capture
    # this closure, a local dictionary must be passed instead.
    exec(func_body, globals(), local_attrs)

    # Return this wrapper method.
    return local_attrs['callable_cached_arged']


@type_check
def property_cached(property_method: CallableTypes) -> PropertyType:
    '''
    Decorate the passed property method to cache the value returned by the
    first implicit call of this method.

    On the first access of a property decorated with this decorator, the passed
    method implementing this property is called, the value returned by this
    property is internally cached into a private attribute of the object to
    which this method is bound, and this value is returned. On each subsequent
    access of this property, this cached value is returned as is *without*
    calling this method. Hence, this method is called at most once for each
    object exposing this property.

    Caveats
    ----------
    **This decorator should always be preferred over the standard**
    :func:`functools.cached_property` decorator available under Python >= 3.8.
    This decorator is substantially more efficient in both space and time than
    that decorator -- which is, of course, the entire point of caching.

    **This decorator does not destroy bound property methods.** Technically,
    the most efficient means of caching a property value into an instance is to
    replace the property method currently bound to that instance with an
    instance variable initialized to that value (e.g., as documented by this
    `StackOverflow answer`_).

    Since a property should only ever be treated as an instance variable,
    there superficially exists little harm in dynamically changing the type of
    the former to the latter. Sadly, doing so introduces numerous subtle issues
    with no plausible workaround.

    In particular, replacing property methods by instance variables:

    * Permits callers to erroneously set **read-only properties** (i.e.,
      properties lacking setter methods), a profound violation of one of the
      principle use cases for properties.
    * Prevents pickling logic elsewhere from automatically excluding cached
      property values, forcing these values to *always* be pickled to disk.
      This is bad. Cached property values are *always* safely recreatable in
      memory (and hence need *not* be pickled) and typically space-consumptive
      in memory (and hence best *not* pickled). The slight efficiency gain from
      replacing property methods by instance variables is hardly worth the
      significant space loss from pickling these variables.

    .. _StackOverflow answer:
        https://stackoverflow.com/a/36684652/2809027
    '''

    # Name of the private instance variable to which this decorator caches the
    # value returned by the decorated property method.
    property_var_name = (
        PROPERTY_CACHED_VAR_NAME_PREFIX + property_method.__name__)

    # Raw string of Python statements comprising the body of this wrapper,
    # including (in order):
    #
    # * A "@wraps" decorator propagating the name, docstring, and other
    #   identifying metadata of the original function to this wrapper.
    # * A private "__property_method" parameter initialized to this function.
    #   In theory, the "property_method" parameter passed to this decorator
    #   should be accessible as a closure-style local in this wrapper. For
    #   unknown reasons (presumably, a subtle bug in the exec() builtin), this
    #   is not the case. Instead, a closure-style local must be simulated by
    #   passing the "property_method" parameter to this function at function
    #   definition time as the default value of an arbitrary parameter.
    #
    # While there exist numerous alternative implementations for caching
    # properties, the approach implemented below has been profiled to be the
    # most efficient. Alternatives include (in order of decreasing efficiency):
    #
    # * Dynamically getting and setting a property-specific key-value pair of
    #   the internal dictionary for the current object, timed to be
    #   approximately 1.5 times as slow as exception handling: e.g.,
    #
    #     if not {property_name!r} in self.__dict__:
    #         self.__dict__[{property_name!r}] = __property_method(self)
    #     return self.__dict__[{property_name!r}]
    #
    # * Dynamically getting and setting a property-specific attribute of
    #   the current object (e.g., the internal dictionary for the current
    #   object), timed to be approximately 1.5 times as slow as exception
    #   handling: e.g.,
    #
    #     if not hasattr(self, {property_name!r}):
    #         setattr(self, {property_name!r}, __property_method(self))
    #     return getattr(self, {property_name!r})
    func_body = '''
@wraps(__property_method)
def property_method_cached(self, __property_method=__property_method):
    try:
        return self.{property_var_name}
    except AttributeError:
        self.{property_var_name} = __property_method(self)
        return self.{property_var_name}
'''.format(property_var_name=property_var_name)

    # Dictionary mapping from local attribute names to values. For efficiency,
    # only attributes required by the body of this wrapper are copied from the
    # current namespace. (See below.)
    local_attrs = {'__property_method': property_method}

    # Dynamically define this wrapper as a closure of this decorator. For
    # obscure and presumably uninteresting reasons, Python fails to locally
    # declare this closure when the locals() dictionary is passed; to capture
    # this closure, a local dictionary must be passed instead.
    exec(func_body, globals(), local_attrs)

    # Return this wrapper method wrapped by a property descriptor.
    return property(local_attrs['property_method_cached'])



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Endoplasmic reticulum (ER) functionality.
'''

# ....................{ IMPORTS                           }....................
import copy
import numpy as np
import matplotlib.pyplot as plt
from betse.science import sim_toolbox as stb
from betse.util.path import dirs, pathnames

# ....................{ CLASSES                           }....................
class EndoRetic(object):
    '''
    Endoplasmic reticulum (ER) modeller.

    This object models all physical phenomena pertaining to the ER, including
    ER-specific pumps, channels, and calcium dynamics (e.g., calcium-induced
    calcium release controlled by inositol-triphosphate). This modeller also
    contains the facilities to initialize, define the core computations for a
    simulation loop, remove the ER during a cutting event, and save and report
    on relevant data.

    Attributes
    ----------
    Ca_er_time : list
        Two-dimensional list of all cellular endoplasmic reticulum calcium ion
        concentrations for all time steps, whose:

        #. First dimension indexes each sampled time step.
        #. Second dimension indexes each cell such that each item is the
           concentration of calcium ions in that cell's endoplasmic reticulum.
    ver_time : list
        Two-dimensional list of all cellular endoplasmic reticulum Vmems
        (transmembrane voltages) for all time steps, whose:

        #. First dimension indexes each sampled time step.
        #. Second dimension indexes each cell such that each item is the
           Vmem (transmembrane voltage) across that cell's endoplasmic
           reticulum.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, sim, cells, p):

        # init basic fields
        self.er_vol = 0.1*cells.cell_vol     # er volume
        self.er_sa = 1.0*cells.cell_sa      # er surface areas
        self.Ver = np.zeros(sim.cdl)   # initial transmembrane voltage for er
        self.Q = np.zeros(sim.cdl)     # total charge in ER
        self.cm_er = p.cm    # ER membrane capacitance

        sim.cc_er = copy.deepcopy(sim.cc_cells)    # ion concentrations
        sim.cc_er[sim.iCa][:] = 0.2                # initial concentration in the ER
        self.Dm_er = copy.deepcopy(sim.cc_cells)    # membrane permeability

        # Membrane permeability altered so all are minimal.
        for arr in self.Dm_er:
            arr[:] = 1.0e-18

        self.Dm_er_base = copy.deepcopy(self.Dm_er)  # copies of Dm for ion channel dynamics
        self.Dm_channels = copy.deepcopy(self.Dm_er)

        self.zer = copy.deepcopy(sim.cc_cells)

        for i, arr in enumerate(self.zer):
            arr[:] = sim.zs[i]

    # ..................{ GETTERS                           }..................
    def get_v(self, sim, p):

        self.Q = np.sum(self.zer*p.F*sim.cc_er, axis = 0)
        self.Ver = (1/self.cm_er)*self.Q*(self.er_vol/self.er_sa)

    # ..................{ CACHERS                           }..................
    def clear_cache(self):

        self.ver_time = []
        self.Ca_er_time = []


    def write_cache(self, sim):

        self.ver_time.append(1*self.Ver)
        self.Ca_er_time.append(1*sim.cc_er[sim.iCa][:])

    # ..................{ UPDATERS                          }..................
    def update(self, sim, cells, p):

        # if p.run_sim:

        self.channels(sim, cells, p)

        # run SERCA pump:
        f_CaATP = stb.pumpCaER(sim.cc_er[sim.iCa], sim.cc_cells[sim.iCa], self.Ver, sim.T, p)

        # update with flux
        sim.cc_cells[sim.iCa] = sim.cc_cells[sim.iCa] - f_CaATP * (self.er_sa / cells.cell_vol) * p.dt
        sim.cc_er[sim.iCa] = sim.cc_er[sim.iCa] + f_CaATP * (self.er_sa / self.er_vol) * p.dt

        for i in sim.movingIons:

            IdCM = np.ones(sim.cdl)

            f_ED = stb.electroflux(sim.cc_cells[i], sim.cc_er[i], self.Dm_er[i], p.tm*IdCM, sim.zs[i]*IdCM,
                self.Ver, sim.T, p, rho=1)

            # update with flux
            sim.cc_cells[i] = sim.cc_cells[i] - f_ED*(self.er_sa/cells.cell_vol)*p.dt
            sim.cc_er[i] = sim.cc_er[i] + f_ED*(self.er_sa/self.er_vol)*p.dt

        self.get_v(sim, p)

        # print(1e3*self.Ver.mean(), sim.cc_er[sim.iCa].mean(), sim.cc_cells[sim.iCa].mean())


    def channels(self, sim, cells, p):

        # Dm_mod_mol = self.gating_max_val * tb.hill(sim.cc_cells[sim.iCa], self.gating_Hill_K, self.gating_Hill_n)
        cCa_act = (sim.cc_cells[sim.iCa]/p.act_Km_Ca)**p.act_n_Ca
        cCa_inh = (sim.cc_cells[sim.iCa] /p.inh_Km_Ca)**p.inh_n_Ca

        Dm_mod_mol = (cCa_act/(1+cCa_act))*(1/(1+cCa_inh))


        if sim.molecules is not None or sim.metabo is not None or sim.grn is not None:

            if sim.molecules is not None and 'IP3' in sim.molecules.molecules:
                cIP3_act = (sim.molecules.cell_concs['IP3']/p.act_Km_IP3)**p.act_n_IP3

            elif sim.metabo is not None and 'IP3' in sim.metabo.core.cell_concs:
                cIP3_act = (sim.metabo.core.cell_concs['IP3'] / p.act_Km_IP3) ** p.act_n_IP3

            elif sim.grn is not None and 'IP3' in sim.grn.core.cell_concs:
                cIP3_act = (sim.grn.core.cell_concs['IP3'] / p.act_Km_IP3) ** p.act_n_IP3

            else:
                cIP3_act = np.zeros(sim.cdl)

            Dm_mod_mol = (cCa_act/(1 + cCa_act))*(1/(1 + cCa_inh))*(cIP3_act/(1+cIP3_act))

        self.Dm_channels[sim.iCa] = p.max_er*Dm_mod_mol

        self.Dm_er = self.Dm_er_base + self.Dm_channels

        # print(sim.cc_er[sim.iCa][p.visual.single_cell_index], sim.cc_cells[sim.iCa][p.visual.single_cell_index],
        # sim.molecules.IP3.c_cells[p.visual.single_cell_index])


    def remove_ers(self, sim, target_inds_cell):

        # remove cells from the mit voltage list:
        ver2 = np.delete(self.Ver, target_inds_cell)
        # reassign the new data vector to the object:
        self.Ver = ver2

        erv2 = np.delete(self.er_vol, target_inds_cell)
        self.er_vol = erv2

        erca2 = np.delete(self.er_sa, target_inds_cell)
        self.er_sa = erca2

        Q2 = np.delete(self.Q, target_inds_cell)
        self.Q = Q2

        cm2 = np.delete(self.cm_er, target_inds_cell)
        self.cm_er = cm2

        # cc_er2 = []
        #
        # for i, arr in enumerate(sim.cc_er):
        #
        #     # remove cells from the mit ion array in sim:
        #     arr2 = np.delete(arr, target_inds_cell)
        #     cc_er2.append(arr2)
        #
        # sim.cc_er = np.asarray(cc_er2)

        der1 = []
        der2 = []
        der3 = []
        zer = []

        for i, arr in enumerate(self.Dm_er):

            arr2 = np.delete(arr, target_inds_cell)
            der1.append(arr2)

        self.Dm_er = np.asarray(der1)

        for i, arr in enumerate(self.Dm_er_base):

            arr2 = np.delete(arr, target_inds_cell)
            der2.append(arr2)

        self.Dm_er_base = np.asarray(der2)

        for i, arr in enumerate(self.Dm_channels):

            arr2 = np.delete(arr, target_inds_cell)
            der3.append(arr2)

        self.Dm_channels = np.asarray(der3)

        for i, arr in enumerate(self.zer):

            arr2 = np.delete(arr, target_inds_cell)
            zer.append(arr2)

        self.zer = np.asarray(zer)



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.


import math

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation


# FIXME to integrate with BETSE, needs to be able to take in forcing function data (which comes from BETSE) and may
# need to be interpolated to this particular grid (along with 'c', by the way, which might be 0 outside of cell cluster)

# FIXME thinking about implementing the FitzHugh-Nagumo or van der Poll relaxation oscillator math as as simple model
# to demonstrate how phase oscillators might create Pilot Wave fields that go on to control the phase oscillators.
# Would want to compare to fMRI images of symmetric brain activity zones.

class WaveSolver(object):

    """
    WaveSolver provides a suite of tools to initialize and iterate a damped, forced wave function on a rectangular
    grid using the Finite Difference Time Domain method.

    Methods
    ---------



    Attributes
    ----------

    """

    def __init__(self,p,constant_c = True):

        # define parameters (these will eventually be linked to the parameters object p)
        self.c = 1     # wave speed [m/s] (can be an array of values defined over the rectangular grid)

        self.xmin = 0  # bounding dimensions of gridded space [m]
        self.xmax = 5
        self.ymin = 0
        self.ymax = 5

        self.extent_p = [self.xmin,self.xmax,self.ymin,self.ymax]  # list holding the world extents

        self.grid_xy_n = 150  # grid points in the x and y dimensions
        self.grid_t_n = 2500  # number of time points

        self.gamma = 1.0e-3  # damping factor [N s/m]

        self.delta_xy = (self.xmax/self.grid_xy_n)   # spacing in the x-y grid

        self.t_end = 20  # end time for simulation

        self.dt = (self.t_end/self.grid_t_n)  # spacing in time grid

        self.x_grid = np.linspace(self.xmin,self.xmax,self.grid_xy_n)
        self.y_grid = np.linspace(self.ymin,self.ymax,self.grid_xy_n)

        self.X, self.Y = np.meshgrid(self.x_grid,self.y_grid)  # gridded X and Y dimensions [m]

        self.fxy = np.exp(-10*((self.X-(self.xmax/2))**2 + (self.Y-(self.ymax/2))**2))  # forcing function amplitude

        f = 0.75 # forcing function frequency

        self.forcing_function = lambda t: np.cos(2*math.pi*f*t)

        self.K = (self.dt*self.c)/self.delta_xy  # Courant number (magnitude relates to stability)

        if self.K > 0.5:

            print('Warning! Your simulation is indicated to be unstable. '
                  'Increase number of time steps or decrease wave speed.')

        self.mShape = list(self.X.shape)
        self.mShape.append(self.grid_t_n)
        self.mShape = tuple(self.mShape)

        self.bounds = 'reflective'   # state what boundary condition applies to the box boundaries

        self.colormap = cm.coolwarm

    def init_matrices(self):  # FIXME next version of this must be able to take in an arbitrary F property to force U
        '''
        Initialize the main matrices of the simulation. Assumes that the initial intensity and velocity of the
        wave property U are zero, and that the substratum is influenced by a forcing function (or property) F.

        '''

        # initial conditions and first time step -- forced damped wave equation
        self.U = np.zeros(self.mShape)   # initialize the U matrix to hold all time steps

        self.F = np.zeros(self.mShape)   # initialize the forcing function F matrix to hold all time steps

        self.F[:,:,0] = self.fxy*self.forcing_function(0)  # create the first time-point of the F matrix

        self.U[:,:,0] = np.zeros(self.X.shape)    # the initial state of the U-property will be zero

        # the next state of the U-property will depend on the forcing function state

        self.U[:,:,1] = ((self.dt**2)/(2-self.dt*self.gamma))*self.F[:,:,0]

    def iter_wave(self,k,t):

        """
        Iterates the wave function under the forcing condition F for time t.

        Parameters
        -----------
        k           Index into the time vector
        t           Time value at index k

        """

        # Advance the forcing function (FIXME later versions will accept forcing *data*)

        self.F[:,:,k] = self.fxy*self.forcing_function(t)

        UU = self.U*self.c  # if c varies over space, this will allow the function to take account for that

        # Interior points:

        self.U[1:-1,1:-1,k+1] = (1/(1+self.gamma))*(self.gamma*self.U[1:-1,1:-1,k]+
                            ((self.dt/self.delta_xy)**2)*(UU[2:,1:-1,k] +
                            UU[0:-2,1:-1,k] + UU[1:-1,2:,k] + UU[1:-1,0:-2,k] - 4*UU[1:-1,1:-1,k]) +
                            2*self.U[1:-1,1:-1,k] - self.U[1:-1,1:-1,k-1] + (self.dt**2)*self.F[1:-1,1:-1,k])


        # Boundary conditions:

        # Dirchlet (fixed boundary, say 0)
        if self.bounds == 'fixed':

            self.U[:,0,k+1] = 0
            self.U[:,-1,k+1] = 0
            self.U[0,:,k+1] = 0
            self.U[-1,:,k+1] = 0


        # reflective boundary
        if self.bounds == 'reflective':

            aa = self.dt/self.delta_xy

            # left boundary, j = 0:
            self.U[1:-1,0,k+1] = (aa**2)*(self.UU[2:,0,k] + self.UU[0:-2,0,k] + 2*self.UU[1:-1,1,k] -
                                              4*self.UU[1:-1,0,k]) + 2*self.U[1:-1,0,k] - self.U[1:-1,0,k-1]

            # right boundary, j = jmax:
            self.U[1:-1,-1,k+1] = (aa**2)*(self.UU[2:,-1,k] + self.UU[0:-2,-1,k] + 2*self.UU[1:-1,-2,k] -
                                               4*self.UU[1:-1,-1,k]) + 2*self.U[1:-1,-1,k] - self.U[1:-1,-1,k-1]

            # bottom boundary, i = 0:
            self.U[0,1:-1,k+1] = (aa**2)*(2*self.UU[1,1:-1,k] + self.UU[0,2:,k] + self.UU[0,0:-2,k] -
                                              4*self.UU[0,1:-1,k]) + 2*self.U[0,1:-1,k] - self.U[0,1:-1,k-1]

            # top boundary, i = max:
            self.U[-1,1:-1,k+1] = (aa**2)*(2*self.UU[-2,1:-1,k] + self.UU[-1,2:,k] + self.UU[-1,0:-2,k] -
                                               4*self.UU[-1,1:-1,k]) + 2*self.U[-1,1:-1,k] - self.U[-1,1:-1,k-1]

        if self.bounds == 'open':

            # left boundary, j = 0:
            self.U[1:-1,0,k+1] = self.U[1:-1,1,k] + ((self.K-1)/(self.K+1))*(self.U[1:-1,1,k+1] - self.U[1:-1,0,k])

            # right boundary, j = jmax:
            self.U[1:-1,-1,k+1] = self.U[1:-1,-2,k] + ((1-self.K)/(1+self.K))*(self.U[1:-1,-1,k] - self.U[1:-1,-2,k+1])


            # bottom boundary, i = 0:
            self.U[0,1:-1,k+1] =  self.U[1,1:-1,k] + ((self.K-1)/(self.K+1))*(self.U[1,1:-1,k+1] - self.U[0,1:-1,k])

            # top boundary, i = max:
            self.U[-1,1:-1,k+1] = self.U[-2,1:-1,k] + ((1-self.K)/(1+self.K))*(self.U[-1,1:-1,k] - self.U[-2,1:-1,k+1])


class AnimateWaves(object):

    def __init__(self,sim, autoscale=True,min_val = -0.50,max_val=0.5):

        fig = plt.figure()

        data = sim.U

        if autoscale is True:

            data_max = np.max(sim.U)
            data_min = np.min(sim.U)

        else:
            data_max = max_val
            data_min = min_val

        l = plt.imshow(sim.U[:,:,0],origin='lower',extent=sim.extent_p,vmin=data_min,vmax=data_max,cmap=sim.colormap)
        plt.colorbar()

        plt.xlim(sim.xmin, sim.xmax)
        plt.ylim(sim.ymin, sim.ymax)
        plt.xlabel('x')
        plt.title('test')
        line_ani = animation.FuncAnimation(fig, self.update_line, sim.grid_t_n, fargs=(data, l),
            interval=50, blit=True)

    def update_line(num, data, line):
        line.set_data(data[:,:,num])
        return line




#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Facilities guaranteeing backward compatibility with prior file formats for
simulation configurations.
'''

#FIXME: Consider generalizing the upgrade_sim_conf() function to preserve all
#in-memory changes back to disk -- albeit, for safety, presumably in a
#separate file preserving the existing filetype and basename prefix of the
#original file (e.g., writing a new "sim_conf-auto_upgraded.yaml" file given
#an input "sim_conf.yaml" file). If this file already exists, silently
#overwrite it; there is absolutely no chance and hence concern of a user
#intentionally embedding the substring "-auto_upgraded" in a simulation
#configuration filename.
#
#Actually, as a safety check, ensure that the current input filename does *NOT*
#already contain this substring. Trivially accomplished, thankfully.

# ....................{ IMPORTS                           }....................
from betse.science.parameters import Parameters
from betse.util.io.log import logs
from betse.util.type.iterable import iterables, iterget
from betse.util.type.obj.sentinels import SENTINEL
from betse.util.type.types import type_check, MappingType

# ....................{ UPGRADERS                         }....................
@type_check
def upgrade_sim_conf(p: Parameters) -> None:
    '''
    Upgrade the in-memory contents of the passed simulation configuration to
    reflect the newest structure of these contents expected by the current
    version of this application.

    This function preserves backward compatibility with all prior supported
    configuration file formats, converting all obsolete key-value pairs of this
    configuration into their modern equivalents. Specifically:

    * Any configuration file produced by any version of this application no
      older than (i.e., at least as new as) the version specified by the
      :attr:`betse.metadata.GIT_TAG_COMPAT_OLDEST` string global is explicitly
      supported by this function and hence guaranteed to be safely loadable
      with the current version of this application.
    * No configuration files produced by any older version of this application
      is explicitly supported by this function. Indeed, these files are
      unlikely to be safely loadable with the current version of this
      application. These files *must* be manually upgraded by end users to
      conform with the current configuration format.
    '''

    #FIXME: Excise this hack *AFTER* refactoring the codebase to use the
    #preferable "p._config" attribute defined above. Since the
    #yaml_alias() data descriptor expects that private attribute rather
    #than this public attribute, this public attribute is unhelpful now.
    p.config = p._conf

    # Upgrade this configuration to each successive format. For safety, each
    # upgrade is performed in strict chronological order.
    _upgrade_sim_conf_to_0_5_0(p)
    _upgrade_sim_conf_to_0_5_2(p)
    _upgrade_sim_conf_to_0_6_0(p)
    _upgrade_sim_conf_to_0_7_1(p)
    _upgrade_sim_conf_to_1_0_0(p)

# ....................{ UPGRADERS ~ 0.5.x                 }....................
@type_check
def _upgrade_sim_conf_to_0_5_0(p: Parameters) -> None:
    '''
    Upgrade the in-memory contents of the passed simulation configuration to
    reflect the newest structure of these contents expected by version 0.5.0
    (i.e., "Happy Hodgkin") of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation configuration to 0.5.0 format...')

    # Localize configuration subdictionaries for convenience.
    results_dict = p._conf['results options']

    # For backward compatibility, convert the prior into the current
    # configuration format.
    if not (
        'while solving' in results_dict and
        'after solving' in results_dict and
        'save' in results_dict
    ):
        # Log a non-fatal warning.
        logs.log_warning(
            'Config file results options '
            '"while solving", "after solving", and/or "save" not found. '
            'Repairing to preserve backward compatibility. '
            'Consider upgrading to the newest config file format!',
        )

        # For convenience, localize configuration subdictionaries.
        anim_save = results_dict['save animations']
        anim_save_frames = anim_save['frames']

        # Convert the prior into the current configuration format.
        results_dict['while solving'] = {
            'animations': {
                'enabled': (
                            results_dict['plot while solving'] or
                            results_dict['save solving plot']
                ),
                'show':    results_dict['plot while solving'],
                'save':    results_dict['save solving plot'],
            },
        }
        results_dict['after solving'] = {
            'plots': {
                'enabled': (
                            results_dict['display plots'] or
                            results_dict['automatically save plots']
                ),
                'show':    results_dict['display plots'],
                'save':    results_dict['automatically save plots'],
            },
            'animations': {
                'enabled': results_dict['create all animations'],
                'show':    results_dict['display plots'],
                'save':    anim_save_frames['enabled'],
            },
        }
        results_dict['save'] = {
            'plots': {
                'filetype': anim_save_frames['filetype'],
                'dpi':      anim_save_frames['dpi'],
            },
            'animations': {
                'images': {
                    'enabled':  anim_save_frames['enabled'],
                    'filetype': anim_save_frames['filetype'],
                    'dpi':      anim_save_frames['dpi'],
                },
                'video': {
                    'enabled':  False,
                    'filetype': 'mkv',
                    'dpi': 300,
                    'bitrate': 1500,
                    'framerate': 5,
                    'metadata': {
                        'artist':  'BETSE',
                        'genre':   'Bioinformatics',
                        'subject': 'Bioinformatics',
                        'comment': 'Produced by BETSE.',
                    },
                    'writers': [
                        'ffmpeg', 'avconv', 'mencoder', 'imagemagick'],
                    'codecs': ['auto'],
                },
            },
            'data': {
                'all': {
                    'enabled': results_dict['export data to file'],
                    'filetype': 'csv',
                },
                'vmem': {
                    'enabled': results_dict['export 2D data to file'],
                    'filetype': 'csv',
                },
            }
        }

    after_solving_anims = results_dict['after solving']['animations']
    after_solving_plots = results_dict['after solving']['plots']

    if 'pipeline' not in after_solving_anims:
        # Log a non-fatal warning.
        logs.log_warning(
            'Config file setting "results options" -> "after solving" -> '
            '"animations" -> "pipeline" not found. '
            'Repairing to preserve backward compatibility. '
            'Consider upgrading to the newest config file format!',
        )

        # Default the value for this dictionary key to the empty list.
        after_solving_anims['pipeline'] = []

    while_solving_anims = results_dict['while solving']['animations']

    if 'colorbar' not in while_solving_anims:
        # Log a non-fatal warning.
        logs.log_warning(
            'Config file setting "results options" -> "while solving" -> '
            '"animations" -> "colorbar" not found. '
            'Repairing to preserve backward compatibility. '
            'Consider upgrading to the newest config file format!',
        )

        # Default the value for this dictionary key to the typical settings.
        while_solving_anims['colorbar'] = {
            'autoscale': True,
            'minimum': -70.0,
            'maximum':  10.0,
        }

    if 'single cell pipeline' not in after_solving_plots:
        # Log a non-fatal warning.
        if 'single cell' not in after_solving_plots:
            logs.log_warning(
                'Config file setting "results options" -> "after solving" -> '
                '"plots" -> "single cell pipeline" not found. '
                'Repairing to preserve backward compatibility. '
                'Consider upgrading to the newest config file format!',
            )

        # Default the value for this dictionary key to the empty list.
        after_solving_plots['single cell pipeline'] = (
            after_solving_plots['single cell']['pipeline']
            if 'single cell' in after_solving_plots else [])

    if 'cell cluster pipeline' not in after_solving_plots:
        # Log a non-fatal warning.
        if 'cell cluster' not in after_solving_plots:
            logs.log_warning(
                'Config file setting "results options" -> "after solving" -> '
                '"plots" -> "cell cluster pipeline" not found. '
                'Repairing to preserve backward compatibility. '
                'Consider upgrading to the newest config file format!',
            )

        after_solving_plots['cell cluster pipeline'] = (
            after_solving_plots['cell cluster']['pipeline']
            if 'cell cluster' in after_solving_plots else [])

    if 'plot networks single cell' not in results_dict:
        # Log a non-fatal warning.
        logs.log_warning(
            'Config file setting "results options" -> '
            '"plot networks single cell" not found. '
            'Repairing to preserve backward compatibility. '
            'Consider upgrading to the newest config file format!',
        )

        # Default the value for this dictionary key to the empty list.
        results_dict['plot networks single cell'] = results_dict[
            'plot single cell graphs']

    # For each pipelined animation and cell cluster plot...
    for anim_conf in iterables.iter_items(
        results_dict['after solving']['animations']['pipeline'],
        results_dict['after solving']['plots']['cell cluster pipeline'],
    ):
        # Add the "enabled" boolean.
        if 'enabled' not in anim_conf:
            anim_conf['enabled'] = True

        # For disambiguity, rename:
        #
        # * "polarization" to "voltage_polarity".
        # * "junction_state" to "gj_permeability".
        if anim_conf['type'] == 'polarization':
            anim_conf['type'] = 'voltage_polarity'
        elif anim_conf['type'] == 'junction_state':
            anim_conf['type'] = 'gj_permeability'


@type_check
def _upgrade_sim_conf_to_0_5_2(p: Parameters) -> None:
    '''
    Upgrade the in-memory contents of the passed simulation configuration to
    reflect the newest structure of these contents expected by version 0.5.2
    (i.e., "Happiest Hodgkin") of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation configuration to 0.5.2 format...')

    # Localize configuration subdictionaries for convenience.
    general_dict = p._conf['general options']
    results_dict = p._conf['results options']
    world_dict = p._conf['world options']

    # Support newly introduced networks plotting.
    if 'plot networks' not in results_dict:
        results_dict['plot networks'] = False
    if 'network colormap' not in results_dict:
        results_dict['network colormap'] = 'coolwarm'

    # Patch old- to new-style cell lattice types.
    if world_dict['lattice type'] == 'hexagonal':
        world_dict['lattice type'] = 'hex'
    elif world_dict['lattice type'] == 'rect':
        world_dict['lattice type'] = 'square'

    # Patch old- to new-style ion profile types.
    if general_dict['ion profile'] == 'basic_Ca':
        general_dict['ion profile'] = 'basic_ca'
    elif general_dict['ion profile'] == 'animal':
        general_dict['ion profile'] = 'mammal'
    elif general_dict['ion profile'] == 'xenopus':
        general_dict['ion profile'] = 'amphibian'
    elif general_dict['ion profile'] == 'customized':
        general_dict['ion profile'] = 'custom'

# ....................{ UPGRADERS ~ 0.6.x                 }....................
@type_check
def _upgrade_sim_conf_to_0_6_0(p: Parameters) -> None:
    '''
    Upgrade the in-memory contents of the passed simulation configuration to
    reflect the newest structure of these contents expected by version 0.6.0
    of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation configuration to 0.6.0 format...')

    # Localize configuration subdictionaries for convenience.
    tissue_dict = p._conf['tissue profile definition']
    vars_dict = p._conf['variable settings']

    # Split ambiguously unified tissue and cut profiles into unambiguous lists.
    if 'profiles' in tissue_dict:
        tissue_dict['cut profiles'] = []
        tissue_dict['tissue profiles'] = []

        for profile in tissue_dict['profiles']:
            if profile['type'] == 'cut':
                tissue_dict['cut profiles'].append(profile)
            else:
                tissue_dict['tissue profiles'].append(profile)

    # Shift the tissue profiles list into a nested dictionary key.
    if 'tissue profiles' in tissue_dict:
        tissue_dict['tissue'] = {
            'profiles': tissue_dict['tissue profiles']
        }

    # Shift the default tissue profile into the tissue profile subsection.
    if 'default' not in tissue_dict['tissue']:
        tissue_dict['tissue']['default'] = {
            'name': vars_dict['default tissue name'],
            'insular': False,
            'diffusion constants': vars_dict['default tissue properties'],
        }

    # If double layer permittivity is undefined, define a sensible default.
    if 'double layer permittivity' not in p._conf['internal parameters']:
        p._conf['internal parameters']['double layer permittivity'] = 80.0

    # Shift the clipping bitmap mask into the default tissue profile.
    if 'clipping' in tissue_dict:
        tissue_dict['tissue']['default']['cell targets'] = tissue_dict[
            'clipping']

    # For disambiguity, upgrade tissue and cut profiles as follows:
    #
    # * Rename "bitmap" to "image" and reduce "image" from a dictionary to
    #   string by promoting its sole key-value pair to itself.
    # * Rename "random" to "percent".
    if 'image' not in tissue_dict['tissue']['default']:
        tissue_dict['tissue']['default']['image'] = tissue_dict[
            'tissue']['default']['cell targets']['bitmap']
    if isinstance(tissue_dict['tissue']['default']['image'], MappingType):
        tissue_dict['tissue']['default']['image'] = tissue_dict[
            'tissue']['default']['image']['file']
    # logs.log_debug('Default tissue: %r', tissue_dict['tissue']['default'])

    for profile in tissue_dict['tissue']['profiles']:
        if profile['cell targets']['type'] == 'bitmap':
            profile['cell targets']['type'] = 'image'
        elif profile['cell targets']['type'] == 'random':
            profile['cell targets']['type'] = 'percent'

        if 'image' not in profile['cell targets']:
            profile['cell targets']['image'] = profile[
                'cell targets']['bitmap']
        if 'percent' not in profile['cell targets']:
            profile['cell targets']['percent'] = profile[
                'cell targets']['random']
        if isinstance(profile['cell targets']['image'], MappingType):
            profile['cell targets']['image'] = profile[
                'cell targets']['image']['file']

    for profile in tissue_dict['cut profiles']:
        if 'image' not in profile:
            profile['image'] = profile['bitmap']
        if isinstance(profile['image'], MappingType):
            profile['image'] = profile['image']['file']

# ....................{ UPGRADERS ~ 0.7.x                 }....................
@type_check
def _upgrade_sim_conf_to_0_7_1(p: Parameters) -> None:
    '''
    Upgrade the in-memory contents of the passed simulation configuration to
    reflect the newest structure of these contents expected by version 0.7.1
    of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation configuration to 0.7.1 format...')

    # Localize configuration subdictionaries for convenience.
    grn_dict     = p._conf['gene regulatory network settings']
    results_dict = p._conf['results options']

    # If the solver type is undefined, default to the complete BETSE solver.
    p._conf.setdefault('fast solver', False)

    # If solver settings are undefined, synthesize from existing defaults.
    p._conf.setdefault('solver options', {
        'type': 'fast' if p._conf['fast solver'] else 'full'
    })

    # If the CSV file pipeline is undefined, synthesize from existing defaults.
    if 'csvs' not in results_dict['after solving']:
        results_dict['after solving']['csvs'] = {
            'save': True,
            'pipeline': [],
        }

        # If the single-cell time series CSV file is enabled, preserve.
        if results_dict['save']['data']['all']['enabled']:
            results_dict['after solving']['csvs']['pipeline'].append({
                'type': 'cell_series', 'enabled': True,})

        # If the all-cells Vmem time series CSV files are enabled, preserve.
        if results_dict['save']['data']['vmem']['enabled']:
            results_dict['after solving']['csvs']['pipeline'].append({
                'type': 'cells_vmem', 'enabled': True,})

    # If CSV save settings are undefined, synthesize from existing defaults.
    results_dict['save'].setdefault('csvs', {'filetype': 'csv',})

    # If "sim-grn" settings are undefined, synthesize from existing defaults.
    grn_dict.setdefault('sim-grn settings', {
        'run network on': 'seed',
        'save to directory': 'RESULTS/GRN',
        'save to file': 'GRN_1.betse.gz',
        'load from': None,
    })

    # If newer "sim-grn" settings are undefined, synthesize from existing
    # defaults.
    grn_dict['sim-grn settings'].setdefault('time step', 0.1)
    grn_dict['sim-grn settings'].setdefault('total time', 1.0e2)
    grn_dict['sim-grn settings'].setdefault('sampling rate', 1.0e1)
    grn_dict['sim-grn settings'].setdefault('run as sim', False)

    # If a "sim-grn" setting previously erroneously defaulting to the string
    # "None" rather than the singleton "None" is set, coerce this setting from
    # the former to the latter.
    if grn_dict['sim-grn settings']['load from'] == 'None':
        grn_dict['sim-grn settings']['load from'] = None

# ....................{ UPGRADERS ~ 0.9.x                 }....................
@type_check
def _upgrade_sim_conf_to_1_0_0(p: Parameters) -> None:
    '''
    Upgrade the in-memory contents of the passed simulation configuration to
    reflect the newest structure of these contents expected by version 1.0.0
    of this application.
    '''

    # Log this upgrade attempt.
    logs.log_debug('Upgrading simulation configuration to 1.0.0 format...')

    # Localize configuration subdictionaries for convenience.
    results_dict = p._conf['results options']
    tissue_dict = p._conf['tissue profile definition']

    # For each tissue profile, define the "color" cell targets type if needed.
    for profile in tissue_dict['tissue']['profiles']:
        if 'color' not in profile['cell targets']:
            profile['cell targets']['color'] = 'ff0000'  # Red. Just 'cause.

    # Define a default uniquified name for each pipelined export.
    _upgrade_sim_conf_to_1_0_0_exports_name(p)

    # If the "visuals" subsection is undefined, define this subsection.
    if 'visuals' not in results_dict:
        results_dict['visuals'] = {
            'cell indices': {
                'show':        results_dict['enumerate cells'],
                'single cell': results_dict['plot cell index'],
            }
        }


@type_check
def _upgrade_sim_conf_to_1_0_0_exports_name(p: Parameters) -> None:
    '''
    Define a default uniquified name for each **pipelined export** (e.g., CSV,
    plot, animation) defined by the passed simulation configuration to reflect
    assumptions expected by version 1.0.0 of this application.
    '''

    # Localize configuration subdictionaries for convenience.
    results_dict = p._conf['results options']

    # Tuple of all export pipelines defined by this configuration.
    export_pipelines = (
        results_dict['after solving']['animations']['pipeline'],
        results_dict['after solving']['csvs']['pipeline'],
        results_dict['after solving']['plots']['cell cluster pipeline'],
        results_dict['after solving']['plots']['single cell pipeline'],
    )

    # First pipelined export (e.g., CSV, plot, animation) defined by this
    # simulation configuration if at least one such pipeline is non-empty *OR*
    # the sentinel placeholder otherwise.
    export_first = iterget.get_item_first_or_sentinel(*export_pipelines)

    # If either all such pipelines are empty *OR* this export is already named,
    # silently reduce to a noop to reduce time complexity.
    #
    # In the former case, we simplistically assume *ALL* pipelined exports to
    # be similarly named. Declaring uniquified names for pipelined exports is
    # highly inefficient and hence *MUST* be avoided for the common case of
    # newer simulation configurations already defining these names.
    if export_first is SENTINEL or 'name' in export_first:
        return
    # Else, this export is *NOT* named. Let's do this.

    # Format specifier for the name of the current pipelined export.
    export_name_format = 'Export ({})'

    # For each export pipeline...
    for export_pipeline in export_pipelines:
        # For each export defined by this pipeline...
        for export in export_pipeline:
            # If this export is unnamed, uniquely name this export.
            if 'name' not in export:
                export['name'] = iterget.get_item_str_uniquified(
                    iterable=export_pipeline,
                    item_key='name',
                    item_str_format=export_name_format,

                    # Internally fallback to a sane name rather than raise
                    # exceptions on unnamed exports in this pipeline.
                    is_defaultable=True,
                )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.util.math.geometry.geopoly` submodule.
'''

# ....................{ IMPORTS                            }....................

# ....................{ TESTS                              }....................
def test_orient_counterclockwise() -> None:
    '''
    Unit test the
    :func:`betse.util.math.geometry.geopoly.orient_counterclockwise`
    function.
    '''

    # Defer heavyweight imports.
    from betse.util.math.geometry.polygon import geopoly

    # Rectangle oriented randomly rather than counter-clockwise. Due to trivial
    # limitations of the betse.lib.numpy.nparray.to_iterable() function, this
    # rectangle is defined as a list of lists rather than tuple of tuples.
    rectangle_unoriented = [[2, 2], [2, -1], [-1, -1], [-1, 2],]

    # Rectangle oriented counter-clockwise.
    rectangle_oriented = geopoly.orient_counterclockwise(rectangle_unoriented)

    # Assert this rectangle to be oriented counter-clockwise.
    assert rectangle_oriented == [[-1, -1], [2, -1], [2, 2], [-1, 2],]

    # Assert that attempting to reorient this rectangle counter-clockwise is a
    # noop (i.e., produces the same rectangle).
    assert geopoly.orient_counterclockwise(rectangle_oriented) == (
        rectangle_oriented)
